import json
import requests
import time
import datetime
import uuid
import urllib
import hmac
import base64
import logging
import tempfile
import pprint
from typing import List

import cloudlanguagetools.service
import cloudlanguagetools.options
import cloudlanguagetools.constants
import cloudlanguagetools.languages
import cloudlanguagetools.ttsvoice
import cloudlanguagetools.constants
import cloudlanguagetools.languages
import cloudlanguagetools.ttsvoice
import cloudlanguagetools.errors
from cloudlanguagetools.options import AudioFormat

logger = logging.getLogger(__name__)

ALIBABA_VOICE_SPEED_DEFAULT = 0
ALIBABA_VOICE_PITCH_DEFAULT = 0

VOICE_OPTIONS = {
    'speed': {'type': 'number_int', 'min': -500, 'max': 500, 'default': ALIBABA_VOICE_SPEED_DEFAULT},
    'pitch': {'type': 'number_int', 'min': -500, 'max': 500, 'default': ALIBABA_VOICE_PITCH_DEFAULT},
    'volume': {'type': 'number_int', 'min': 0, 'max': 100, 'default': 50},
    cloudlanguagetools.options.AUDIO_FORMAT_PARAMETER: {
        'type': cloudlanguagetools.options.ParameterType.list.name,
        'values': ['pcm', 'wav', 'mp3'],
        'default': 'mp3'
    }
}


class AlibabaService(cloudlanguagetools.service.Service):
    def __init__(self):
        self.access_token = None
        
    def configure(self, config):
        self.access_key_id = config['access_key_id']
        self.access_key_secret = config['access_key_secret']
        self.app_key = config['app_key']

    def refresh_token(self):
        logger.info("refreshing token")
        params = {
            "AccessKeyId": self.access_key_id,
            "Action": "CreateToken",
            "Version": "2019-07-17",
            "Format": "JSON",
            "RegionId": "ap-southeast-1",
            "Timestamp": datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ"),
            "SignatureMethod": "HMAC-SHA1",
            "SignatureVersion": "1.0",
            "SignatureNonce": str(uuid.uuid4())
        }

        # sort by keys alphabetically
        params = dict(sorted(params.items()))
        # timestamp needs to be double-quoted by the end
        params["Timestamp"] = urllib.parse.quote(params["Timestamp"], safe='')
        
        # urlencode with noop lambda for no quoting - we will quote later
        params_str = urllib.parse.urlencode(params, quote_via=lambda a, b, c, d: a)
        
        # this is always /, as we always hit the path / on the API
        url_encoded = urllib.parse.quote("/", safe='')
        str_to_sign = f"GET&{url_encoded}&{urllib.parse.quote(params_str, safe='')}"
        str_to_sign = str_to_sign.encode("utf-8")
        
        key = self.access_key_secret + "&"
        key = key.encode("utf-8")
        
        # calculate HMAC-SHA1 digest and convert to base64
        dig = hmac.new(key, str_to_sign, "sha1").digest()
        dig = base64.standard_b64encode(dig).decode("utf-8")
        
        # signature also needs to be quoted
        signature = urllib.parse.quote(dig, safe='')
        params_str = f"Signature={signature}&{params_str}"
        
        response = requests.get(f"http://nlsmeta.ap-southeast-1.aliyuncs.com/?{params_str}")
        
        if response.status_code != 200:
            logger.warning(f"Token request failed: {response.text}")
            raise cloudlanguagetools.errors.RequestError("Token request failed", None, response.text)
            
        data = response.json()
        self.access_token = data["Token"]
        logger.info(f"Got access token: {self.access_token}")

    def get_tts_audio(self, text, voice_key, options):
        logger.debug(f'get_tts_audio, text: {text} voice_key: {voice_key}')
        if not self.access_token or self.access_token["ExpireTime"] <= int(time.time()):
            self.refresh_token()

        speed = int(options.get('speed', ALIBABA_VOICE_SPEED_DEFAULT))
        pitch = int(options.get('pitch', ALIBABA_VOICE_PITCH_DEFAULT))
        voice = voice_key['name']

        params = {
            "format": "mp3",
            "appkey": self.app_key,
            "speech_rate": speed,
            "pitch_rate": pitch,
            "text": text,
            "token": self.access_token["Id"],
            "voice": voice
        }

        response = requests.get(
            "https://nls-gateway-ap-southeast-1.aliyuncs.com/stream/v1/tts",
            params=params,
            timeout=cloudlanguagetools.constants.RequestTimeout
        )

        logger.debug(f'response status code: {response.status_code} headers: {pprint.pformat(response.headers)} length of response: {len(response.content)}')

        if response.status_code != 200:
            data = response.json()
            error_message = f'could not retrieve audio using Alibaba: {response.content}'
            logger.error(error_message)
            raise cloudlanguagetools.errors.RequestError(text, voice, error_message)

        if response.headers['Content-Type'] != 'audio/mpeg':
            logger.error(f'Unexpected response type. Response as text: {response.text}')
            raise cloudlanguagetools.errors.RequestError(
                text, voice,
                f'Got bad content type in response: {response.headers["Content-Type"]}'
            )
        
        # Create a temporary file and write the audio content to it
        output_temp_file = tempfile.NamedTemporaryFile(prefix=f'cloudlanguage_tools_{self.__class__.__name__}_audio', suffix='.mp3')
        output_temp_filename = output_temp_file.name
        with open(output_temp_filename, 'wb') as audio:
            audio.write(response.content)        
        logger.debug(f'result temporary file name: {output_temp_file.name}')
        return output_temp_file

    def get_tts_voice_list(self):
        # returns list of TtsVoice
        return []

    def build_tts_voice_v3(self, name: str, voice_id: str, gender: cloudlanguagetools.constants.Gender, audio_languages: List[cloudlanguagetools.languages.AudioLanguage]):
        return cloudlanguagetools.ttsvoice.TtsVoice_v3(
            name=name,
            voice_key={'name': voice_id},
            options=VOICE_OPTIONS,
            service=cloudlanguagetools.constants.Service.Alibaba,
            gender=gender,
            audio_languages=audio_languages,
            service_fee=cloudlanguagetools.constants.ServiceFee.paid
        )

    def get_tts_voice_list_v3(self) -> List[cloudlanguagetools.ttsvoice.TtsVoice_v3]:
        result = []

        # Standard voices for all scenarios
        result.extend([
            # Standard Chinese voices
            self.build_tts_voice_v3('Xiaoyun (Standard)', 'Xiaoyun', 
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN, cloudlanguagetools.languages.AudioLanguage.en_US]),
            self.build_tts_voice_v3('Xiaogang (Standard)', 'Xiaogang',
                cloudlanguagetools.constants.Gender.Male, [cloudlanguagetools.languages.AudioLanguage.zh_CN, cloudlanguagetools.languages.AudioLanguage.en_US]),
            self.build_tts_voice_v3('Ruoxi (Gentle)', 'Ruoxi',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN, cloudlanguagetools.languages.AudioLanguage.en_US]),
            self.build_tts_voice_v3('Siqi (Gentle)', 'Siqi',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN, cloudlanguagetools.languages.AudioLanguage.en_US]),
            self.build_tts_voice_v3('Sijia (Standard)', 'Sijia',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN, cloudlanguagetools.languages.AudioLanguage.en_US]),
            self.build_tts_voice_v3('Sicheng (Standard)', 'Sicheng',
                cloudlanguagetools.constants.Gender.Male, [cloudlanguagetools.languages.AudioLanguage.zh_CN, cloudlanguagetools.languages.AudioLanguage.en_US]),
            self.build_tts_voice_v3('Aiqi (Gentle)', 'Aiqi',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN, cloudlanguagetools.languages.AudioLanguage.en_US]),
            self.build_tts_voice_v3('Aijia (Standard)', 'Aijia',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN, cloudlanguagetools.languages.AudioLanguage.en_US]),
            self.build_tts_voice_v3('Aicheng (Standard)', 'Aicheng',
                cloudlanguagetools.constants.Gender.Male, [cloudlanguagetools.languages.AudioLanguage.zh_CN, cloudlanguagetools.languages.AudioLanguage.en_US]),
            self.build_tts_voice_v3('Aida (Standard)', 'Aida',
                cloudlanguagetools.constants.Gender.Male, [cloudlanguagetools.languages.AudioLanguage.zh_CN, cloudlanguagetools.languages.AudioLanguage.en_US]),
            self.build_tts_voice_v3("Ning'er", 'Ninger',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN]),
            self.build_tts_voice_v3('Ruilin', 'Ruilin',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN]),

            # Customer service voices
            self.build_tts_voice_v3('Siyue (Gentle)', 'Siyue',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN, cloudlanguagetools.languages.AudioLanguage.en_US]),
            self.build_tts_voice_v3('Aiya (Harsh)', 'Aiya',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN, cloudlanguagetools.languages.AudioLanguage.en_US]),
            self.build_tts_voice_v3('Aixia (Amiable)', 'Aixia',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN, cloudlanguagetools.languages.AudioLanguage.en_US]),
            self.build_tts_voice_v3('Aimei (Sweet)', 'Aimei',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN, cloudlanguagetools.languages.AudioLanguage.en_US]),
            self.build_tts_voice_v3('Aiyu (Natural)', 'Aiyu',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN, cloudlanguagetools.languages.AudioLanguage.en_US]),
            self.build_tts_voice_v3('Aiyue (Gentle)', 'Aiyue',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN, cloudlanguagetools.languages.AudioLanguage.en_US]),
            self.build_tts_voice_v3('Aijing (Harsh)', 'Aijing',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN, cloudlanguagetools.languages.AudioLanguage.en_US]),
            self.build_tts_voice_v3('Xiaomei (Sweet)', 'Xiaomei',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN, cloudlanguagetools.languages.AudioLanguage.en_US]),

            # Regional accent voices
            self.build_tts_voice_v3('Aina (Zhejiang)', 'Aina',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN]),
            self.build_tts_voice_v3('Yina (Zhejiang)', 'Yina',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN]),
            self.build_tts_voice_v3('Sijing (Harsh)', 'Sijing',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN]),

            # Child voices
            self.build_tts_voice_v3('Sitong (Child)', 'Sitong',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN]),
            self.build_tts_voice_v3('Xiaobei (Little Girl)', 'Xiaobei',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN]),
            self.build_tts_voice_v3('Aitong (Child)', 'Aitong',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN]),
            self.build_tts_voice_v3('Aiwei (Little Girl)', 'Aiwei',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN]),
            self.build_tts_voice_v3('Aibao (Little Girl)', 'Aibao',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN]),

            # English voices
            self.build_tts_voice_v3('Harry (British)', 'Harry',
                cloudlanguagetools.constants.Gender.Male, [cloudlanguagetools.languages.AudioLanguage.en_GB]),
            self.build_tts_voice_v3('Abby (American)', 'Abby',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.en_US]),
            self.build_tts_voice_v3('Andy (American)', 'Andy',
                cloudlanguagetools.constants.Gender.Male, [cloudlanguagetools.languages.AudioLanguage.en_US]),
            self.build_tts_voice_v3('Eric (British)', 'Eric',
                cloudlanguagetools.constants.Gender.Male, [cloudlanguagetools.languages.AudioLanguage.en_GB]),
            self.build_tts_voice_v3('Emily (British)', 'Emily',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.en_GB]),
            self.build_tts_voice_v3('Luna (British)', 'Luna',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.en_GB]),
            self.build_tts_voice_v3('Luca (British)', 'Luca',
                cloudlanguagetools.constants.Gender.Male, [cloudlanguagetools.languages.AudioLanguage.en_GB]),
            self.build_tts_voice_v3('Wendy (British)', 'Wendy',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.en_GB]),
            self.build_tts_voice_v3('William (British)', 'William',
                cloudlanguagetools.constants.Gender.Male, [cloudlanguagetools.languages.AudioLanguage.en_GB]),
            self.build_tts_voice_v3('Olivia (British)', 'Olivia',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.en_GB]),

            # Special dialect voices
            self.build_tts_voice_v3('Shanshan (Cantonese)', 'Shanshan',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_HK]),
            self.build_tts_voice_v3('Xiaoyue (Sichuan)', 'Xiaoyue',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN]),
            self.build_tts_voice_v3('Lydia (Bilingual)', 'Lydia',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.en_US]),
            self.build_tts_voice_v3('Aishuo (Natural)', 'Aishuo',
                cloudlanguagetools.constants.Gender.Male, [cloudlanguagetools.languages.AudioLanguage.zh_CN, cloudlanguagetools.languages.AudioLanguage.en_US]),
            self.build_tts_voice_v3('Qingqing (Taiwanese)', 'Qingqing',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN]),
            self.build_tts_voice_v3('Cuijie (Northeastern)', 'Cuijie',
                cloudlanguagetools.constants.Gender.Female, [cloudlanguagetools.languages.AudioLanguage.zh_CN]),
            self.build_tts_voice_v3('Xiaoze (Hunan)', 'Xiaoze',
                cloudlanguagetools.constants.Gender.Male, [cloudlanguagetools.languages.AudioLanguage.zh_CN])
        ])
        
        return result



import os
import sys
import logging
import unittest
import pprint
import requests
import time
import json

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import cloudlanguagetools
import cloudlanguagetools.servicemanager
from cloudlanguagetools.languages import Language
from cloudlanguagetools.constants import Service
import cloudlanguagetools.errors

def get_manager():
    manager = cloudlanguagetools.servicemanager.ServiceManager()
    manager.configure_default()
    return manager

class TestBreakdown(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        logging.basicConfig(format='%(asctime)s %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s', 
                            datefmt='%Y%m%d-%H:%M:%S',
                            stream=sys.stdout,
                            level=logging.DEBUG)

        cls.manager = get_manager()

        num_tries = 3
        success = False
        while success == False and num_tries >= 0:
            num_tries -= 1
            try:
                cls.language_data = cls.manager.get_language_data_json()
                success = True
            except requests.exceptions.ReadTimeout as e:
                logging.exception(f'could not get language data, timeout')
                time.sleep(1)

        

    def test_language_list(self):
        self.assertTrue(len(self.language_data['language_list']) > 0)
        # check for the presence of a few languages
        self.assertEqual('French', self.language_data['language_list']['fr'])
        self.assertEqual('Chinese (Simplified)', self.language_data['language_list']['zh_cn'])
        self.assertEqual('Thai', self.language_data['language_list']['th'])
        self.assertEqual('Chinese (Cantonese, Traditional)', self.language_data['language_list']['yue'])
        self.assertEqual('Chinese (Traditional)', self.language_data['language_list']['zh_tw'])    

    def test_breakdown_thai(self):
        # pytest test_breakdown.py -s -rPP -k test_breakdown_thai
        # pytest test_breakdown.py -s --log-cli-level=DEBUG -k test_breakdown_thai

        text = 'ดิฉันอายุยี่สิบเจ็ดปีค่ะ'
        source_language = 'th'
        target_language = 'en'

        # choose breakdown
        tokenization_service = 'PyThaiNLP'
        tokenization_candidates = [x for x in self.language_data['tokenization_options'] if x['language_code'] == source_language and x['service'] == tokenization_service]
        self.assertEqual(len(tokenization_candidates), 1)
        tokenization_option = tokenization_candidates[0]

        # choose transliteration
        transliteration_service = 'PyThaiNLP'
        transliteration_candidates = [x for x in self.language_data['transliteration_options'] if x['language_code'] == source_language and x['service'] == transliteration_service]
        transliteration_option = transliteration_candidates[0]

        # choose translation
        translation_service = 'Azure'
        source_language_id = [x for x in self.language_data['translation_options'] if x['language_code'] == source_language and x['service'] == translation_service][0]['language_id']
        target_language_id = [x for x in self.language_data['translation_options'] if x['language_code'] == target_language and x['service'] == translation_service][0]['language_id']
        translation_option = {
            'service': translation_service,
            'source_language_id': source_language_id,
            'target_language_id': target_language_id
        }

        # run the breakdown
        breakdown_result = self.manager.get_breakdown(text, tokenization_option, translation_option, transliteration_option)
        pprint.pprint(breakdown_result)

        expected_output = [{'lemma': 'ดิฉัน',
        'token': 'ดิฉัน',
        'translation': 'I',
        'transliteration': 'dichan'},
        {'lemma': 'อายุ',
        'token': 'อายุ',
        'translation': 'age',
        'transliteration': 'ayu'},
        {'lemma': 'ยี่สิบ',
        'token': 'ยี่สิบ',
        'translation': 'twenty',
        'transliteration': 'yisip'},
        {'lemma': 'เจ็ด',
        'token': 'เจ็ด',
        'translation': 'seven',
        'transliteration': 'chet'},
        {'lemma': 'ปี', 'token': 'ปี', 'translation': 'year', 'transliteration': 'pi'},
        {'lemma': 'ค่ะ',
        'token': 'ค่ะ',
        'translation': 'yes',
        'transliteration': 'kha'}]

        self.assertEqual(breakdown_result, expected_output)

        # try same breakdown with spaces, result should be the same
        text = 'ดิฉัน อายุ ยี่สิบเจ็ดปีค่ะ'
        breakdown_result = self.manager.get_breakdown(text, tokenization_option, translation_option, transliteration_option)
        self.assertEqual(breakdown_result, expected_output)

        # add punctuation
        text = 'ดิฉัน, อายุ ยี่สิบเจ็ดปีค่ะ'
        breakdown_result = self.manager.get_breakdown(text, tokenization_option, translation_option, transliteration_option)
        self.assertEqual(breakdown_result, expected_output)        


    def test_breakdown_english(self):
        # pytest test_breakdown.py -s -rPP -k test_breakdown_english

        text = "I was reading today's paper."
        source_language = 'en'
        target_language = 'fr'

        # choose breakdown
        tokenization_service = 'Spacy'
        tokenization_candidates = [x for x in self.language_data['tokenization_options'] if x['language_code'] == source_language and x['service'] == tokenization_service]
        self.assertEqual(len(tokenization_candidates), 1)
        tokenization_option = tokenization_candidates[0]

        # choose transliteration
        transliteration_service = 'Epitran'
        transliteration_candidates = [x for x in self.language_data['transliteration_options'] if x['language_code'] == source_language and x['service'] == transliteration_service]
        transliteration_option = transliteration_candidates[0]

        # choose translation
        translation_service = 'Azure'
        source_language_id = [x for x in self.language_data['translation_options'] if x['language_code'] == source_language and x['service'] == translation_service][0]['language_id']
        target_language_id = [x for x in self.language_data['translation_options'] if x['language_code'] == target_language and x['service'] == translation_service][0]['language_id']
        translation_option = {
            'service': translation_service,
            'source_language_id': source_language_id,
            'target_language_id': target_language_id
        }

        # run the breakdown
        breakdown_result = self.manager.get_breakdown(text, tokenization_option, translation_option, transliteration_option)
        pprint.pprint(breakdown_result)

        expected_output = [{'lemma': 'I',
        'pos_description': 'pronoun, personal',
        'token': 'I',
        'translation': 'Je',
        'transliteration': 'aj'},
        {'lemma': 'be',
        'pos_description': 'verb, past tense',
        'token': 'was',
        'translation': 'être',
        'transliteration': 'wɑz'},
        {'lemma': 'read',
        'pos_description': 'verb, gerund or present participle',
        'token': 'reading',
        'translation': 'lire',
        'transliteration': 'ɹɛdɪŋ'},
        {'lemma': 'today',
        'pos_description': 'noun, singular or mass',
        'token': 'today',
        'translation': 'Aujourd’hui',
        'transliteration': 'tədej'},
        {'lemma': "'s", 'pos_description': 'possessive ending', 'token': "'s"},
        {'lemma': 'paper',
        'pos_description': 'noun, singular or mass',
        'token': 'paper',
        'translation': 'papier',
        'transliteration': 'pejpɹ̩'},
        {'lemma': '.',
        'pos_description': 'punctuation mark, sentence closer',
        'token': '.'}]

        self.assertEqual(breakdown_result, expected_output)

    def test_tokenization_pythainlp(self):
        # pytest test_breakdown.py -rPP -k test_tokenization_pythainlp

        service = cloudlanguagetools.constants.Service.PyThaiNLP.name

        tokenization_candidates = [x for x in self.language_data['tokenization_options'] if x['language_code'] == Language.th.name and x['service'] == service]
        self.assertEqual(len(tokenization_candidates), 1)
        tokenization_option = tokenization_candidates[0]

        # thai
        tokenization_result = self.manager.get_tokenization('ดิฉันอายุยี่สิบเจ็ดปีค่ะ', service, tokenization_option['tokenization_key'])
        
        expected_result = [
            {'token': 'ดิฉัน', 'lemma': 'ดิฉัน', 'can_translate': True, 'can_transliterate': True}, 
            {'token': 'อายุ', 'lemma': 'อายุ', 'can_translate': True, 'can_transliterate': True},
            {'token': 'ยี่สิบ', 'lemma': 'ยี่สิบ', 'can_translate': True, 'can_transliterate': True},
            {'token': 'เจ็ด', 'lemma': 'เจ็ด', 'can_translate': True, 'can_transliterate': True},
            {'token': 'ปี', 'lemma': 'ปี', 'can_translate': True, 'can_transliterate': True},
            {'token': 'ค่ะ', 'lemma': 'ค่ะ', 'can_translate': True, 'can_transliterate': True}
        ]

        self.assertEqual(tokenization_result, expected_result)

    def test_tokenization_spacy(self):
        # pytest test_breakdown.py -rPP -k test_tokenization_spacy

        service = cloudlanguagetools.constants.Service.Spacy.name

        language = Language.en.name
        tokenization_candidates = [x for x in self.language_data['tokenization_options'] if x['language_code'] == language and x['service'] == service]
        self.assertEqual(len(tokenization_candidates), 1)
        tokenization_option = tokenization_candidates[0]

        # english
        # =======

        text = "I was reading today's paper."
        tokenization_result = self.manager.get_tokenization(text, service, tokenization_option['tokenization_key'])
        # pprint.pprint(tokenization_result)
        
        expected_result = [{'can_translate': True,
            'can_transliterate': True,
            'lemma': 'I',
            'pos_description': 'pronoun, personal',
            'token': 'I'},
            {'can_translate': True,
            'can_transliterate': True,
            'lemma': 'be',
            'pos_description': 'verb, past tense',
            'token': 'was'},
            {'can_translate': True,
            'can_transliterate': True,
            'lemma': 'read',
            'pos_description': 'verb, gerund or present participle',
            'token': 'reading'},
            {'can_translate': True,
            'can_transliterate': True,
            'lemma': 'today',
            'pos_description': 'noun, singular or mass',
            'token': 'today'},
            {'can_translate': False,
            'can_transliterate': False,
            'lemma': "'s",
            'pos_description': 'possessive ending',
            'token': "'s"},
            {'can_translate': True,
            'can_transliterate': True,
            'lemma': 'paper',
            'pos_description': 'noun, singular or mass',
            'token': 'paper'},
            {'can_translate': False,
            'can_transliterate': False,
            'lemma': '.',
            'pos_description': 'punctuation mark, sentence closer',
            'token': '.'}]
        self.assertEqual(tokenization_result, expected_result)        

    def test_tokenization_spacy_french(self):
        # pytest test_breakdown.py -rPP -k test_tokenization_spacy_french

        self.maxDiff = None

        service = cloudlanguagetools.constants.Service.Spacy.name

        # french
        # ======

        language = Language.fr.name
        tokenization_candidates = [x for x in self.language_data['tokenization_options'] if x['language_code'] == language and x['service'] == service]
        self.assertEqual(len(tokenization_candidates), 1)
        tokenization_option = tokenization_candidates[0]

        text = "Le nouveau plan d’investissement du gouvernement."
        tokenization_result = self.manager.get_tokenization(text, service, tokenization_option['tokenization_key'])
        # pprint.pprint(tokenization_result)
        
        expected_result = [{'can_translate': True,
            'can_transliterate': True,
            'lemma': 'le',
            'pos_description': 'determiner',
            'token': 'Le'},
            {'can_translate': True,
            'can_transliterate': True,
            'lemma': 'nouveau',
            'pos_description': 'adjective',
            'token': 'nouveau'},
            {'can_translate': True,
            'can_transliterate': True,
            'lemma': 'plan',
            'pos_description': 'noun',
            'token': 'plan'},
            {'can_translate': False,
            'can_transliterate': False,
            'lemma': 'd’',
            'pos_description': 'adposition',
            'token': 'd’'},
            {'can_translate': True,
            'can_transliterate': True,
            'lemma': 'investissement',
            'pos_description': 'noun',
            'token': 'investissement'},
            {'can_translate': True,
            'can_transliterate': True,
            'lemma': 'de',
            'pos_description': 'adposition',
            'token': 'du'},
            {'can_translate': True,
            'can_transliterate': True,
            'lemma': 'gouvernement',
            'pos_description': 'noun',
            'token': 'gouvernement'},
            {'can_translate': False,
            'can_transliterate': False,
            'lemma': '.',
            'pos_description': 'punctuation',
            'token': '.'}]

        self.assertEqual(tokenization_result, expected_result)                


    def test_tokenization_spacy_chinese(self):
        # pytest test_breakdown.py -rPP -k test_tokenization_spacy_chinese

        service = cloudlanguagetools.constants.Service.Spacy.name

        language = Language.zh_cn.name
        tokenization_options = [x for x in self.language_data['tokenization_options'] if x['language_code'] == language and x['service'] == service]

        text = "送外卖的人"

        expected_result_chars = [{'can_translate': True,
            'can_transliterate': True,
            'lemma': '送',
            'token': '送'},
            {'can_translate': True,
            'can_transliterate': True,
            'lemma': '外',
            'token': '外'},
            {'can_translate': True,
            'can_transliterate': True,
            'lemma': '卖',
            'token': '卖'},
            {'can_translate': True,
            'can_transliterate': True,
            'lemma': '的',
            'token': '的'},
            {'can_translate': True,
            'can_transliterate': True,
            'lemma': '人',
            'token': '人'}]

        expected_result_words = [{'can_translate': True,
            'can_transliterate': True,
            'lemma': '送',
            'token': '送'},
            {'can_translate': True,
            'can_transliterate': True,
            'lemma': '外卖',
            'token': '外卖'},
            {'can_translate': True,
            'can_transliterate': True,
            'lemma': '的',
            'token': '的'},
            {'can_translate': True,
            'can_transliterate': True,
            'lemma': '人',
            'token': '人'}]

        self.assertEqual(self.manager.get_tokenization(text, service, tokenization_options[0]['tokenization_key']), expected_result_words)
        self.assertEqual(self.manager.get_tokenization(text, service, tokenization_options[1]['tokenization_key']), expected_result_chars)
        self.assertEqual(self.manager.get_tokenization(text, service, tokenization_options[2]['tokenization_key']), expected_result_words)        



# these exceptions can be retried
class TransientError(Exception):
    pass

# no need to retry, something is wrong which won't be fixed on retry
class PermanentError(Exception):
    pass

class InputError(PermanentError):
    pass

class ApiKeyNotFoundError(PermanentError):
    pass

class RequestError(TransientError):
    pass

class TimeoutError(TransientError):
    pass

class NotFoundError(PermanentError):
    pass

class OverQuotaError(PermanentError):
    pass


import wave
import tempfile

# audio utilities


# take PCM audio and wrap it in a WAV container
def wrap_pcm_data_wave(audio_temp_file: tempfile.NamedTemporaryFile,
        num_channels,
        sample_width,
        framerate) -> tempfile.NamedTemporaryFile:
    # read the audio data
    f = open(audio_temp_file.name, 'rb')
    data = f.read()
    f.close()

    wav_frames = []
    wav_frames.append(data)

    wav_audio_temp_file = tempfile.NamedTemporaryFile(prefix='clt_wav_audio_', suffix='.wav')

    WAVEFORMAT = wave.open(wav_audio_temp_file.name,'wb')
    WAVEFORMAT.setnchannels(num_channels) # one channel, mono
    WAVEFORMAT.setsampwidth(sample_width) # Polly's output is a stream of 16-bits (2 bytes) samples
    WAVEFORMAT.setframerate(framerate)
    WAVEFORMAT.writeframes(b''.join(wav_frames))
    WAVEFORMAT.close()    
    
    return wav_audio_temp_file


import os
import sys
import logging
import unittest
import pytest
import json
import pprint
import time
import requests
import backoff

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import cloudlanguagetools
import cloudlanguagetools.servicemanager
from cloudlanguagetools.languages import Language
from cloudlanguagetools.constants import Service
import cloudlanguagetools.errors

def get_manager():
    manager = cloudlanguagetools.servicemanager.ServiceManager()
    manager.configure_default()
    return manager

class TestTranslation(unittest.TestCase):
    
    @classmethod
    @backoff.on_exception(backoff.expo,
                        requests.exceptions.RequestException,
                        max_time=60)
    def get_all_language_data(cls):
        cls.language_list = cls.manager.get_language_list()
        cls.translation_language_list = cls.manager.get_translation_language_list_json()
        cls.transliteration_language_list = cls.manager.get_transliteration_language_list_json()
        cls.tokenization_options = cls.manager.get_tokenization_options_json()

    @classmethod
    def setUpClass(cls):
        cls.manager = get_manager()
        cls.get_all_language_data()

    def test_language_list(self):
        self.assertTrue(len(self.language_list) > 0)
        # check for the presence of a few languages
        self.assertEqual('French', self.language_list['fr'])
        self.assertEqual('Chinese (Simplified)', self.language_list['zh_cn'])
        self.assertEqual('Thai', self.language_list['th'])
        self.assertEqual('Chinese (Cantonese, Traditional)', self.language_list['yue'])
        self.assertEqual('Chinese (Traditional)', self.language_list['zh_tw'])

    def test_detection(self):
        source_text = 'Je ne suis pas intéressé.'
        self.assertEqual(Language.fr, self.manager.detect_language([source_text]))

        source_list = [
        'Pouvez-vous me faire le change ?',
        'Pouvez-vous débarrasser la table, s\'il vous plaît?',
        "I'm not interested"
        ]    
        # french should still win
        self.assertEqual(Language.fr, self.manager.detect_language(source_list))

        # chinese
        self.assertEqual(Language.zh_cn, self.manager.detect_language(['我试着每天都不去吃快餐']))

        # chinese traditional (most cantonese text will be recognized as traditional/taiwan)
        self.assertEqual(Language.zh_tw, self.manager.detect_language(['你住得好近一個機場']))
        

    def test_translate(self):
        source_text = 'Je ne suis pas intéressé.'
        translated_text = self.manager.get_translation(source_text, cloudlanguagetools.constants.Service.Azure.name, 'fr', 'en')
        self.assertEqual(translated_text, "I'm not interested.")

    def test_translate_error(self):
        source_text = 'Je ne suis pas intéressé.'
        self.assertRaises(cloudlanguagetools.errors.RequestError, self.manager.get_translation, source_text, cloudlanguagetools.constants.Service.Azure.name, 'fr', 'zh_cn')

    def sanitize_text(self, recognized_text):
        result_text = recognized_text.replace('.', '').\
            replace('。', '').\
            replace('?', '').\
            replace('？', '').\
            replace('您', '你').\
            replace(':', '').lower()
        return result_text

    def translate_text(self, service, source_text, source_language, target_language, expected_result):
        source_language_list = [x for x in self.translation_language_list if x['language_code'] == source_language.name and x['service'] == service.name]
        self.assertEqual(1, len(source_language_list))
        target_language_list = [x for x in self.translation_language_list if x['language_code'] == target_language.name and x['service'] == service.name]
        self.assertEqual(1, len(target_language_list))

        from_language_key = source_language_list[0]['language_id']
        to_language_key = target_language_list[0]['language_id']

        # now, translate
        translated_text = self.manager.get_translation(source_text, service.name, from_language_key, to_language_key)


        translated_text = self.sanitize_text(translated_text)

        if isinstance(expected_result, list):
            # multiple possible accepted translations
            expected_results = [self.sanitize_text(x) for x in expected_result]
            self.assertIn(translated_text, expected_results)
        else:
            # single value
            expected_result = self.sanitize_text(expected_result)
            self.assertEqual(expected_result, translated_text)

    def test_translate_chinese(self):
        # pytest test_translation.py -k test_translate_chinese
        self.translate_text(Service.Azure, '送外卖的人', Language.zh_cn, Language.en, ['the person who delivers the takeaway', 
        'people who deliver food', 'people who deliver takeaways', 'the person who delivered the takeaway', 'food delivery people',
        'delivery people'])
        self.translate_text(Service.Google, '中国有很多外国人', Language.zh_cn, Language.en, 'There are many foreigners in China')
        self.translate_text(Service.Azure, '成本很低', Language.zh_cn, Language.fr, 'Le coût est faible')
        self.translate_text(Service.Google, '换登机牌', Language.zh_cn, Language.fr, 
            ["Changer la carte d'embarquement", 
             "changer de carte d'embarquement", 
             "changer la carte d'embarquement",
             "échanger la carte d'embarquement"])
        self.translate_text(Service.Amazon, '换登机牌', Language.zh_cn, Language.fr, 
            ["utilisez votre carte d'embarquement", # seems wrong, but amazon returns this occasionally
             "modifier la carte d'embar", # seems wrong, but amazon returns this occasionally
             "carte d'embarquement", 
             "modifier la carte d'embarquement", 
             "changer la carte d'embarquement",
             "échangez votre carte d'embarquement",
             "modifiez votre carte d'embarquement"])

    def test_translate_chinese_watson(self):
        self.translate_text(Service.Watson, '中国有很多外国人', Language.zh_cn, Language.en, 'There are a lot of foreigners in China.')

    def test_translate_chinese_amazon(self):
        self.translate_text(Service.Amazon, '中国有很多外国人', Language.zh_cn, Language.en, ['there are many foreigners in china', 'there are a lot of foreigners in china'])

    def test_translate_naver(self):
        # pytest test_translation.py -k test_translate_naver
        self.translate_text(Service.Naver, '천천히 말해 주십시오', Language.ko, Language.en, 'Please speak slowly.')
        self.translate_text(Service.Naver, 'Please speak slowly', Language.en, Language.ko, 
            ['천천히 말씀해 주세요', 
             '천천히 말해주세요'])

        self.translate_text(Service.Naver, '천천히 말해 주십시오', Language.ko, Language.fr, 
        [ "s'il vous plaît, parlez lentement", 
          "parlez lentement, s'il vous plaît",
          'parlez lentement'])

        self.translate_text(Service.Naver, 'Veuillez parler lentement.', Language.fr, Language.ko, '천천히 말씀해 주세요')        

    @pytest.mark.skip('2022/07 seems to work now')
    def test_translate_naver_unsupported_pair(self):
        # pytest test_translation.py -k test_translate_naver_unsupported_pair
        self.assertRaises(cloudlanguagetools.errors.RequestError, self.translate_text, Service.Naver, 'Veuillez parler lentement.', Language.fr, Language.th, 'Please speak slowly.')

    def test_translate_deepl(self):
        # pytest tests/test_translation.py -rPP -k test_translate_deepl
        self.translate_text(Service.DeepL, 'Please speak slowly', Language.en, Language.fr, 'Veuillez parler lentement')
        self.translate_text(Service.DeepL, 'Je ne suis pas intéressé.', Language.fr, Language.en, ["""I'm not interested.""", 'i am not interested'])
        self.translate_text(Service.DeepL, '送外卖的人', Language.zh_cn, Language.en, ['delivery person', 'takeaway delivery people'])

    def test_translate_portuguese_deepl(self):
        # pytest tests/test_translation.py -rPP -k test_translate_portuguese_deepl
        self.translate_text(Service.DeepL, 'Please speak slowly', Language.en, Language.pt_pt, 'por favor, fale devagar')
        self.translate_text(Service.DeepL, 'Please speak slowly', Language.en, Language.pt_br, 'por favor, fale devagar')

        self.translate_text(Service.DeepL, 'por favor, fale devagar', Language.pt_pt, Language.fr, 'veuillez parler lentement')
        self.translate_text(Service.DeepL, 'por favor, fale devagar', Language.pt_br, Language.fr, 'veuillez parler lentement')        

    
    # 2022/09/13: argos service disabled
    # def test_translate_chinese_argos(self):
    #     self.translate_text(Service.ArgosTranslate, '中国有很多外国人', Language.zh_cn, Language.en, 'there are many foreigners in china')

    # def test_translate_french_argos(self):
    #     self.translate_text(Service.ArgosTranslate, 'Please speak slowly', Language.en, Language.fr, 'Veuillez parler lentement.')

    @pytest.mark.skip('2023/08 libretranslate disabled')
    def test_translate_french_libretranslate(self):
        self.translate_text(Service.LibreTranslate, 'Please speak slowly', Language.en, Language.fr, 'Veuillez parler lentement.')        

    @pytest.mark.skip('2023/08 libretranslate disabled')
    def test_translate_chinese_libretranslate(self):
        self.translate_text(Service.LibreTranslate, '中国有很多外国人', Language.zh_cn, Language.en, 'there are many foreigners in china')

    def test_translate_all(self):
        # pytest test_translation.py -rPP -k test_translate_all
        # pytest test_translation.py --capture=no --log-cli-level=INFO -k test_translate_all

        source_text = '成本很低'
        from_language = Language.zh_cn.name
        to_language =  Language.fr.name
        result = self.manager.get_all_translations(source_text, from_language, to_language)
        self.assertTrue('Azure' in result)
        self.assertTrue('Google' in result)
        self.assertTrue('Watson' in result)

        possible_french_translations = [
            'Très faible coût',
            'Le coût est faible',
            'Le coût est très faible',
            'à bas prix', 
            'Faible coût', 
            'À bas prix', 
            'faible coût', 
            'très faible coût',
            'Le coût est très bas.'
        ]

        self.assertIn(result['Azure'], possible_french_translations)
        self.assertIn(result['Google'], possible_french_translations)
        self.assertIn(result['Watson'], possible_french_translations)

    def test_translate_all_bug_vi(self):
        # pytest test_translation.py -rPP -k test_translate_all_bug_vi
        # pytest test_translation.py --capture=no --log-cli-level=INFO -k test_translate_all

        source_text = '<b><span style="font-weight: 400;">Thấy ghê … sao ám tui hoài vậy?</span></b>'
        from_language = Language.vi.name
        to_language =  Language.en.name
        result = self.manager.get_all_translations(source_text, from_language, to_language)
        pprint.pprint(result)
        self.assertTrue('Azure' in result)
        self.assertTrue('Google' in result)
        self.assertTrue('Watson' in result)
        self.assertTrue
        self.assertEqual(result['Watson'], "<b><span style=\"font-weight: 400;\">What's wrong with you?</span></b>")

    def test_transliteration_azure_custom(self):
        # pytest tests/test_translation.py -k test_transliteration_azure_custom
        
        # chinese
        source_text = '成本很低'
        service = 'Azure'
        # transliteration_key = transliteration_option['transliteration_key']
        transliteration_key = {
            'language_id': 'zh-Hans', 
            'from_script': 'Hans', 
            'to_script': 'Latn',
        }
        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertIn(result, ['chéng běn hěn dī', 'chéngběn hěndī'])

        transliteration_key = {
            'language_id': 'zh-Hant', 
            'from_script': 'Hans', 
            'to_script': 'Hant',
        }
        result = self.manager.get_transliteration('讲话', service, transliteration_key)
        self.assertEqual(result, '講話')

    def test_transliteration_azure(self):
        # pytest tests/test_translation.py -k test_transliteration_azure
        
        # chinese
        source_text = '成本很低'
        from_language = Language.zh_cn.name
        service = 'Azure'
        transliteration_candidates = [
            x for x in self.transliteration_language_list 
            if x['language_code'] == from_language 
            and x['service'] == service
            and 'to Latin' in x['transliteration_shortname']]
        self.assertTrue(len(transliteration_candidates) == 1)
        transliteration_option = transliteration_candidates[0]
        service = transliteration_option['service']
        transliteration_key = transliteration_option['transliteration_key']
        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertIn(result, ['chéng běn hěn dī', 'chéngběn hěndī'])

        # thai
        source_text = 'ประเทศไทย'
        service = 'Azure'
        from_language = Language.th.name
        transliteration_candidates = [x for x in self.transliteration_language_list if x['language_code'] == from_language and x['service'] == service]
        self.assertEqual(len(transliteration_candidates), 1)
        transliteration_option = transliteration_candidates[0]
        service = transliteration_option['service']
        transliteration_key = transliteration_option['transliteration_key']
        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertEqual('prathetthai', result)

    def test_pinyin_jyutping(self):
        # pytest tests/test_translation.py -k test_pinyin_jyutping
        # test direct access to the pinyin and jyuting functions (special case)


        # pinyin
        # ------

        source_text = '成本很低'
        result = self.manager.get_pinyin(source_text, False, False)
        self.assertEqual(result, {'word_list': ['成本', '很', '低'], 'solutions': [['chéngběn'], ['hěn'], ['dī']]})

        # try with corrections
        corrections = [
            {
                'chinese': '低',
                'pinyin': 'di4'
            }
        ]
        result = self.manager.get_pinyin(source_text, False, False, corrections=corrections)
        self.assertEqual(result, {'word_list': ['成本', '很', '低'], 'solutions': [['chéngběn'], ['hěn'], ['dì', 'dī']]})

        # jyutping
        # --------

        source_text = '全身按摩'
        expected_result = {'word_list': ['全身', '按摩'], 'solutions': [['cyùnsān'], ['ônmō']]}
        result = self.manager.get_jyutping(source_text, False, False)
        self.assertEqual(result, expected_result)

        # try with corrections
        corrections = [
            {
                'chinese': '按摩',
                'jyutping': 'on1mo1'
            }
        ]        
        result = self.manager.get_jyutping(source_text, False, False, corrections=corrections)
        self.assertEqual(result, {'word_list': ['全身', '按摩'], 'solutions': [['cyùnsān'], ['ōnmō', 'ônmō']]})


    def test_transliteration_mandarincantonese(self):
        # pytest tests/test_translation.py -k test_transliteration_mandarincantonese
        
        # chinese
        source_text = '成本很低'
        from_language = Language.zh_cn.name
        service = 'MandarinCantonese'
        transliteration_candidates = [x for x in self.transliteration_language_list if x['language_code'] == from_language 
            and x['service'] == service 
            and x['transliteration_key']['tone_numbers'] == False
            and x['transliteration_key']['spaces'] == False]

        self.assertTrue(len(transliteration_candidates) == 1)
        transliteration_option = transliteration_candidates[0]
        service = transliteration_option['service']
        transliteration_key = transliteration_option['transliteration_key']
        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertEqual('chéngběn hěn dī', result)

        # tone numbers
        transliteration_candidates = [x for x in self.transliteration_language_list if x['language_code'] == from_language 
            and x['service'] == service 
            and x['transliteration_key']['tone_numbers'] == True
            and x['transliteration_key']['spaces'] == False]
        self.assertTrue(len(transliteration_candidates) == 1) 
        transliteration_option = transliteration_candidates[0]       
        transliteration_key = transliteration_option['transliteration_key']

        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertEqual('cheng2ben3 hen3 di1', result)

        # jyutping
        # '我出去攞野食'

        source_text = '我出去攞野食'
        from_language = Language.yue.name

        transliteration_candidates = [x for x in self.transliteration_language_list if x['language_code'] == from_language 
            and x['service'] == service 
            and x['transliteration_key']['tone_numbers'] == False
            and x['transliteration_key']['spaces'] == False]

        self.assertTrue(len(transliteration_candidates) == 1)

        transliteration_option = transliteration_candidates[0]
        service = transliteration_option['service']
        transliteration_key = transliteration_option['transliteration_key']
        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertEqual('ngǒ cēothêoi ló jěsik', result)

    def verify_easypronunciation_english(self, input, expected_output):
        service = cloudlanguagetools.constants.Service.EasyPronunciation.name

        source_text = input
        from_language = Language.en.name
        transliteration_candidates = [x for x in self.transliteration_language_list if x['language_code'] == from_language \
            and x['service'] == service \
            and x['transliteration_key'].get('variant', None) == None]
        self.assertTrue(len(transliteration_candidates) == 1)
        transliteration_option = transliteration_candidates[0]
        service = transliteration_option['service']
        transliteration_key = transliteration_option['transliteration_key']
        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertEqual(expected_output, result)


    def test_transliteration_easypronunciation(self):
        # pytest test_translation.py -rPP -k test_transliteration_easypronunciation

        # easypronunciation IPA test
        test_easy_pronunciation = True
        if not test_easy_pronunciation:
            return

        service = cloudlanguagetools.constants.Service.EasyPronunciation.name

        # french
        source_text = 'l’herbe est plus verte ailleurs'
        from_language = Language.fr.name
        transliteration_candidates = [x for x in self.transliteration_language_list if x['language_code'] == from_language and x['service'] == service]
        self.assertTrue(len(transliteration_candidates) == 1)
        transliteration_option = transliteration_candidates[0]
        service = transliteration_option['service']
        transliteration_key = transliteration_option['transliteration_key']
        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertEqual('lɛʁb‿ ɛ ply vɛʁt‿ ajœʁ', result)

        # english
        source_text = 'do you have a boyfriend'
        from_language = Language.en.name
        transliteration_candidates = [x for x in self.transliteration_language_list if x['language_code'] == from_language and x['service'] == service]
        #self.assertTrue(len(transliteration_candidates) == 1)
        transliteration_option = transliteration_candidates[0]
        service = transliteration_option['service']
        transliteration_key = transliteration_option['transliteration_key']
        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertIn(result, ['ˈduː ˈjuː ˈhæv ə ˈbɔɪˌfɹɛnd', 'ˈduː jə həv ə ˈbɔɪˌfɹɛnd'])        

        self.verify_easypronunciation_english('take', 'ˈteɪk')
        self.verify_easypronunciation_english('poor', 'ˈpʊr')
        self.verify_easypronunciation_english('self', 'ˈsɛlf')

        # italian
        source_text = 'Piacere di conoscerla.'
        from_language = Language.it.name
        transliteration_candidates = [x for x in self.transliteration_language_list if x['language_code'] == from_language and x['service'] == service]
        self.assertTrue(len(transliteration_candidates) == 1)
        transliteration_option = transliteration_candidates[0]
        service = transliteration_option['service']
        transliteration_key = transliteration_option['transliteration_key']
        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertIn(result, ['pjaˈtʃere di konoʃʃerla', 'pjaˈtʃere di koˈnoʃʃerla'])

        # japanese - Kana
        source_text = 'おはようございます'
        from_language = Language.ja.name
        transliteration_candidates = [x for x in self.transliteration_language_list if x['language_code'] == from_language and x['service'] == service and 'Kana' in x['transliteration_name']]
        self.assertTrue(len(transliteration_candidates) == 1)
        transliteration_option = transliteration_candidates[0]
        service = transliteration_option['service']
        transliteration_key = transliteration_option['transliteration_key']
        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertEqual('おはよー ございます', result)

        # japanese - romaji
        source_text = 'おはようございます'
        from_language = Language.ja.name
        transliteration_candidates = [x for x in self.transliteration_language_list if x['language_code'] == from_language and x['service'] == service and 'Romaji' in x['transliteration_name']]
        self.assertTrue(len(transliteration_candidates) == 1)
        transliteration_option = transliteration_candidates[0]
        service = transliteration_option['service']
        transliteration_key = transliteration_option['transliteration_key']
        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertEqual('ohayo– gozaimasu', result)

        # portuguese - portugal
        source_text = 'Posso olhar a cozinha?'
        from_language = Language.pt_pt.name
        transliteration_candidates = [x for x in self.transliteration_language_list if x['language_code'] == from_language and x['service'] == service]
        self.assertTrue(len(transliteration_candidates) == 1)
        transliteration_option = transliteration_candidates[0]
        service = transliteration_option['service']
        transliteration_key = transliteration_option['transliteration_key']
        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertEqual('ˈpɔsu oˈʎaɾ ɐ kuˈziɲɐ', result)

        # portuguese - brazil
        source_text = 'Perdi a minha carteira.'
        from_language = Language.pt_br.name
        transliteration_candidates = [x for x in self.transliteration_language_list if x['language_code'] == from_language and x['service'] == service]
        self.assertTrue(len(transliteration_candidates) == 1)
        transliteration_option = transliteration_candidates[0]
        service = transliteration_option['service']
        transliteration_key = transliteration_option['transliteration_key']
        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertEqual('peʁˈdʒi a ˈmiɲɐ kaʁˈtejɾɐ', result)

        # spanish
        source_text = '¿A qué hora usted cierra?'
        from_language = Language.es.name
        transliteration_candidates = [x for x in self.transliteration_language_list if x['language_code'] == from_language and x['service'] == service]
        self.assertTrue(len(transliteration_candidates) == 1)
        transliteration_option = transliteration_candidates[0]
        service = transliteration_option['service']
        transliteration_key = transliteration_option['transliteration_key']
        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertIn(result, ['a ˈke ˈoɾa u̯sˈtɛð ˈsjɛra', 'a ˈke ˈoɾa wsˈteð ˈsjera'])

        # german
        source_text = 'Können Sie mir das auf der Karte zeigen?'
        from_language = Language.de.name
        transliteration_candidates = [x for x in self.transliteration_language_list if x['language_code'] == from_language and x['service'] == service]
        self.assertTrue(len(transliteration_candidates) == 1)
        transliteration_option = transliteration_candidates[0]
        service = transliteration_option['service']
        transliteration_key = transliteration_option['transliteration_key']
        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertEqual('ˈkœ.nən ˈziː ˈmiːɐ̯ das ˈaʊ̯f deːɐ ˈkar.tə ˈtsaɪ̯.ɡn̩', result)        

        # russian
        source_text = 'Могу я посмотреть меню?'
        from_language = Language.ru.name
        transliteration_candidates = [x for x in self.transliteration_language_list if x['language_code'] == from_language and x['service'] == service]
        self.assertTrue(len(transliteration_candidates) == 1)
        transliteration_option = transliteration_candidates[0]
        service = transliteration_option['service']
        transliteration_key = transliteration_option['transliteration_key']
        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertIn(result, ['mɐ.ˈɡu ˈjæ pə.smɐ.ˈtrʲetʲ mʲɪ.ˈnʲʉ', 'mɐ.ˈɡu ˈja pə.smɐ.ˈtrʲetʲ mʲɪ.ˈnʲʉ'])


    def test_transliteration_easypronunciation_english_british_american(self):
        # pytest tests/test_translation.py -rPP -k test_transliteration_easypronunciation_english_british_american

        service = cloudlanguagetools.constants.Service.EasyPronunciation.name
        from_language = Language.en.name

        # british english
        # ===============

        transliteration_candidates = [x for x in self.transliteration_language_list if x['language_code'] == from_language \
            and x['service'] == service \
            and x['transliteration_key'].get('variant', None) == 'British']
        self.assertTrue(len(transliteration_candidates) == 1)
        transliteration_option = transliteration_candidates[0]
        service = transliteration_option['service']
        transliteration_key = transliteration_option['transliteration_key']

        source_text = 'car'
        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertEqual('ˈkɑː', result)
        
        source_text = 'lot'
        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertEqual('ˈlɒt', result)

        source_text = 'aluminium'
        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertEqual('ˌæljəˈmɪniəm', result)

        # american english
        # ================

        transliteration_candidates = [x for x in self.transliteration_language_list if x['language_code'] == from_language \
            and x['service'] == service \
            and x['transliteration_key'].get('variant', None) == 'American']
        self.assertTrue(len(transliteration_candidates) == 1)
        transliteration_option = transliteration_candidates[0]
        service = transliteration_option['service']
        transliteration_key = transliteration_option['transliteration_key']

        source_text = 'car'
        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertEqual('ˈkɑːr', result)
        
        source_text = 'lot'
        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertEqual('ˈlɑːt', result)

        source_text = 'aluminium'
        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertEqual('ˌæljəˈmɪniːəm', result)


    def verify_transliteration(self, source_text, transliteration_option, expected_output):
        service = transliteration_option['service']
        transliteration_key = transliteration_option['transliteration_key']
        result = self.manager.get_transliteration(source_text, service, transliteration_key)
        self.assertEqual(expected_output, result)        


    def verify_transliteration_single_option(self, from_language, source_text, service, expected_result):
        from_language_name = from_language.name
        transliteration_candidates = [x for x in self.transliteration_language_list if x['language_code'] == from_language_name and x['service'] == service]
        self.assertEqual(len(transliteration_candidates), 1)
        transliteration_option = transliteration_candidates[0]
        self.verify_transliteration(source_text, transliteration_option, expected_result)


    def verify_transliteration_multiple_options(self, from_language, source_text, service, expected_result_list):
        from_language_name = from_language.name
        transliteration_candidates = [x for x in self.transliteration_language_list if x['language_code'] == from_language_name and x['service'] == service]

        actual_result_list = []
        for transliteration_option in transliteration_candidates:
            transliteration_key = transliteration_option['transliteration_key']
            result = self.manager.get_transliteration(source_text, service, transliteration_key)
            actual_result_list.append(result)

        # sort both actual and expected
        actual_result_list.sort()
        expected_result_list.sort()

        self.assertEqual(actual_result_list, expected_result_list)


    def test_transliteration_epitran(self):
        # pytest test_translation.py -rPP -k test_transliteration_epitran

        service = cloudlanguagetools.constants.Service.Epitran.name

        # french
        self.verify_transliteration_multiple_options(Language.fr, 'l’herbe est plus verte ailleurs', service, ['l’ɛʀbə ɛst plys vɛʀtə elœʀ', 'l’hɛrbɛ ɛst plys vɛrtɛ elœrs'])

        # english 
        self.verify_transliteration_single_option(Language.en, 'do you have a boyfriend', service, 'du ju hæv ə bojfɹɛnd')

        # german
        self.verify_transliteration_multiple_options(Language.de, 'Können Sie mir das auf der Karte zeigen?', 
            service, 
            ['kønnən siː mir das awf dər karte t͡sajeːɡən?',
            'kønən siː miʁ das auf deʁ kaɐte t͡saieɡən?',
            'kœnɛːn ziː miːr daːs aʊ̯f dɛːr kaːrtɛː t͡saɪ̯ɡɛːn?'])

        # spanish
        self.verify_transliteration_single_option(Language.es, '¿A qué hora usted cierra?', service, '¿a ke oɾa usted siera?')

    def test_transliteration_pythainlp(self):
        # pytest test_translation.py -rPP -k test_transliteration_pythainlp

        service = cloudlanguagetools.constants.Service.PyThaiNLP.name

        # thai
        self.verify_transliteration_multiple_options(Language.th, 'สวัสดี', service, ['s a ˧ . w a t̚ ˨˩ . d iː ˧', 'sawatdi'])




import os
import sys
import logging
import unittest
import pytest
import json
import pprint

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import cloudlanguagetools
import cloudlanguagetools.servicemanager
import cloudlanguagetools.errors

def get_manager():
    manager = cloudlanguagetools.servicemanager.ServiceManager()
    manager.configure_default()
    return manager

class TestTranslation(unittest.TestCase):
    def setUp(self):
        self.manager = get_manager()


    def sanitize_text(self, translation_text):
        result_text = translation_text.replace('.', '').\
            replace(',', '').\
            replace('。', '').\
            replace('?', '').\
            replace('？', '').\
            replace('您', '你').\
            replace(':', '').lower()
        return result_text

    def test_single_prompt(self):
        prompt = 'translate to French: speak slowly please.'
        response, tokens = self.manager.openai_single_prompt(prompt)
        self.assertIn(self.sanitize_text(response),
                      [
                         self.sanitize_text("Parlez lentement s'il vous plaît."),
                         self.sanitize_text("parle lentement s'il te plaît") 
                      ])

    def test_full_query(self):
        # pytest --log-cli-level=DEBUG tests/test_llm.py -k test_full_query
        messages = [
                {"role": "system", "content": "You are a friendly assistant who will translate languages"},
                {"role": "user", "content": "translate into english: 成本很低"},
        ]
        response = self.manager.openai_full_query(messages)
        new_message = response.choices[0].message
        # pprint.pprint(new_message)
        messages.append(new_message)
        messages.append({
            'role': 'user',
            'content': 'take the last message  and translate into French.'
        })
        pprint.pprint(messages)
        response = self.manager.openai_full_query(messages)
        pprint.pprint(response)
        new_message = response.choices[0].message
        pprint.pprint(new_message)
        sanitized_output = self.sanitize_text(new_message.content)
        self.assertTrue(self.sanitize_text('coût') in sanitized_output, f'output: [{sanitized_output}]')
        self.assertTrue(self.sanitize_text('faible') in sanitized_output or self.sanitize_text('bas') in sanitized_output, 
            f'output: [{sanitized_output}]')


import os
import sys
import logging
import unittest
import json
import pytest
import pprint

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

LOAD_TEST_SERVICES_ONLY = os.environ.get('CLOUDLANGUAGETOOLS_CORE_TEST_SERVICES', 'no') == 'yes'

import cloudlanguagetools
import cloudlanguagetools.servicemanager
from cloudlanguagetools.languages import Language
from cloudlanguagetools.constants import Service
import cloudlanguagetools.errors

def get_manager():
    manager = cloudlanguagetools.servicemanager.ServiceManager()
    manager.configure_default()
    return manager

class TestMockServices(unittest.TestCase):
    
    def test_language_data(self):
        if not LOAD_TEST_SERVICES_ONLY:
            pytest.skip('you must set CLOUDLANGUAGETOOLS_CORE_TEST_SERVICES=yes')

        manager = get_manager()
        language_data = manager.get_language_data_json_v2()
        self.assertTrue(len(language_data['premium']) > 0)
        self.assertTrue(len(language_data['free']) > 0)


    def test_translation(self):
        if not LOAD_TEST_SERVICES_ONLY:
            pytest.skip('you must set CLOUDLANGUAGETOOLS_CORE_TEST_SERVICES=yes')

        manager = get_manager()
        language_data = manager.get_language_data_json_v2()
        translation_options = language_data['free']['translation_options']
        # pprint.pprint(translation_options)
        translation_options_fr = [x for x in translation_options if x['language_code'] == 'fr']
        self.assertEqual(len(translation_options_fr), 1)

        translated_text_str = manager.get_translation('text_input', 'TestServiceA', 'fr', 'en')
        translated_text_obj = json.loads(translated_text_str)
        translated_text_expected = {
            'text': 'text_input',
            'from_language_key': 'fr',
            'to_language_key': 'en'
        }
        self.assertEqual(translated_text_obj, translated_text_expected)

        translated_text_str = manager.get_translation('text_input', 'TestServiceB', 'fr', 'en')
        translated_text_obj = json.loads(translated_text_str)
        self.assertEqual(translated_text_obj, translated_text_expected)

    def test_transliteration(self):
        if not LOAD_TEST_SERVICES_ONLY:
            pytest.skip('you must set CLOUDLANGUAGETOOLS_CORE_TEST_SERVICES=yes')

        manager = get_manager()
        language_data = manager.get_language_data_json_v2()
        translation_options = language_data['free']['transliteration_options']
        # pprint.pprint(translation_options)
        translation_options_zh = [x for x in translation_options if x['language_code'] == 'zh_cn']
        self.assertEqual(len(translation_options_zh), 1)

        transliterated_text_str = manager.get_transliteration('text_input', 'TestServiceA', translation_options_zh[0]['transliteration_key'])
        transliterated_text_obj = json.loads(transliterated_text_str)
        transliterated_text_expected = {
            'text': 'text_input',
            'transliteration_key': 'pinyin'
        }
        self.assertEqual(transliterated_text_obj, transliterated_text_expected)

    def test_dictionary_lookup(self):
        if not LOAD_TEST_SERVICES_ONLY:
            pytest.skip('you must set CLOUDLANGUAGETOOLS_CORE_TEST_SERVICES=yes')

        manager = get_manager()
        language_data = manager.get_language_data_json_v2()
        dict_lookup_options = language_data['free']['dictionary_lookup_options']
        # pprint.pprint(translation_options)
        dict_lookup_options_zh = [x for x in dict_lookup_options if x['language_code'] == 'zh_cn']
        self.assertEqual(len(dict_lookup_options_zh), 1)

        transliterated_text_str = manager.get_dictionary_lookup('text_input', 'TestServiceA', dict_lookup_options_zh[0]['lookup_key'])
        transliterated_text_obj = json.loads(transliterated_text_str)
        transliterated_text_expected = {
            'type': 'dictionary',
            'text': 'text_input',
            'lookup_key': 'zh'
        }
        self.assertEqual(transliterated_text_obj, transliterated_text_expected)

    def test_service_cost(self):
        if not LOAD_TEST_SERVICES_ONLY:
            pytest.skip('you must set CLOUDLANGUAGETOOLS_CORE_TEST_SERVICES=yes')        

        manager = get_manager()
        # test services
        self.assertEqual(manager.service_cost('abcd', 'TestServiceA', cloudlanguagetools.constants.RequestType.transliteration), 0)
        self.assertEqual(manager.service_cost('abcd', 'TestServiceB', cloudlanguagetools.constants.RequestType.transliteration), 4)



import unittest
import logging
import random
import requests
import re
import sys
import os
import time
import magic
import pytest
import json
import time
import pprint
import functools
import tempfile
import backoff

import audio_utils

CLOUDLANGUAGETOOLS_CORE_TEST_UNRELIABLE = os.environ.get('CLOUDLANGUAGETOOLS_CORE_TEST_UNRELIABLE', 'no') == 'yes'

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import cloudlanguagetools
import cloudlanguagetools.servicemanager
import cloudlanguagetools.options
import cloudlanguagetools.errors
from cloudlanguagetools.languages import Language
from cloudlanguagetools.languages import AudioLanguage
from cloudlanguagetools.constants import Service

logger = logging.getLogger(__name__)

def get_manager():
    manager = cloudlanguagetools.servicemanager.ServiceManager()
    manager.configure_default()

    return manager

BACKOFF_MAX_TIME=30

def skip_unreliable_clt_test():
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if not CLOUDLANGUAGETOOLS_CORE_TEST_UNRELIABLE:
                pytest.skip(f'you must set CLOUDLANGUAGETOOLS_CORE_TEST_UNRELIABLE=yes')
            return func(*args, **kwargs)
        return wrapper
    return decorator


@backoff.on_exception(backoff.expo,
                      requests.exceptions.RequestException,
                      max_time=BACKOFF_MAX_TIME)
def get_tts_voice_list_json_with_retry(manager):
    return manager.get_tts_voice_list_json()

@backoff.on_exception(backoff.expo,
                      requests.exceptions.RequestException, 
                      max_time=BACKOFF_MAX_TIME)
def get_tts_voice_list_v3_with_retry(manager):
    return manager.get_tts_voice_list_v3()

class TestAudio(unittest.TestCase):

    ENGLISH_INPUT_TEXT = 'This is the best restaurant in town.'
    FRENCH_INPUT_TEXT = 'Bonjour'
    JAPANESE_INPUT_TEXT = 'こんにちは'
    CHINESE_INPUT_TEXT = '你好'
    KOREAN_INPUT_TEXT = '안녕하세요'

    @classmethod
    def setUpClass(cls):
        cls.manager = get_manager()
        cls.language_list = cls.manager.get_language_list()
        cls.voice_list = get_tts_voice_list_json_with_retry(cls.manager)
        cls.voice_list_v3 = get_tts_voice_list_v3_with_retry(cls.manager)
        
        import http.client as http_client                            
        http_client.HTTPConnection.debuglevel = 1
        requests_log = logging.getLogger("requests.packages.urllib3")
        requests_log.setLevel(logging.DEBUG)
        requests_log.propagate = True



    def get_voice_list_for_language(self, language):
        subset = [x for x in self.voice_list if x['language_code'] == language.name]
        return subset

    def get_voice_list_service_audio_language(self, service, audio_language):
        subset = [x for x in self.voice_list if x['audio_language_code'] == audio_language.name and x['service'] == service.name]
        return subset        

    def get_voice_by_service_and_name(self, service: Service, voice_name) -> cloudlanguagetools.ttsvoice.TtsVoice_v3:
        subset = [x for x in self.voice_list_v3 if voice_name in x.name and x.service == service]
        num_voices = len(subset)
        self.assertEqual(num_voices, 1, msg=f'found {num_voices} voices for {service} and {voice_name}, expected 1')
        return subset[0]

    def get_voice_by_lambda(self, service: Service, filter_func, assert_unique=True):
        service_voices = [x for x in self.voice_list_v3 if x.service == service]
        subset = [x for x in service_voices if filter_func(x)]
        if assert_unique:
            self.assertEqual(len(subset), 1, pprint.pformat(subset))
        return subset[0]

    def verify_voice(self, voice, text, recognition_language):
        return self.verify_voice_internal(voice['voice_key'], voice['service'], text, recognition_language)

    def verify_voice_v3(self, voice, text, recognition_language):
        voice_key = voice.voice_key
        voice_service = voice.service.name
        return self.verify_voice_internal(voice_key, voice_service, text, recognition_language)

    @backoff.on_exception(backoff.expo,
                        cloudlanguagetools.errors.TransientError,
                        max_time=BACKOFF_MAX_TIME)
    def get_tts_audio_with_retry(self, text, voice_service, voice_key):
        audio_temp_file = self.manager.get_tts_audio(text, voice_service, voice_key, {})
        return audio_temp_file

    def verify_voice_internal(self, voice_key, voice_service, text, recognition_language):

        audio_temp_file = self.get_tts_audio_with_retry(text, voice_service, voice_key)

        # check file format
        is_mp3 = audio_utils.is_mp3_format(audio_temp_file.name)
        if not is_mp3:
            mime_type = magic.from_file(audio_temp_file.name)
            logger.error(f'found mime type: {mime_type}')
            # dump file to console
            f = open(audio_temp_file.name, "r")
            print(f.read())

        self.assertTrue(is_mp3)

        audio_format = cloudlanguagetools.options.AudioFormat.mp3

        self.recognize_and_verify_text(audio_temp_file, text, recognition_language, audio_format)


    def recognize_and_verify_text(self, 
                audio_temp_file: tempfile.NamedTemporaryFile, 
                expected_text: str,
                recognition_language: str,
                audio_format: cloudlanguagetools.options.AudioFormat):
        # recognize text
        # ==============
        sanitized_expected_text = audio_utils.sanitize_recognized_text(expected_text)
        # first, try openwhisper
        logger.info(f'attempting to recognize text using OpenAI. expected text: {sanitized_expected_text}')
        audio_text_openai = audio_utils.speech_to_text_openai(self.manager, audio_temp_file, audio_format)
        sanitized_openai_text = audio_utils.sanitize_recognized_text(audio_text_openai)
        if sanitized_expected_text == sanitized_openai_text:
            # openai text matches
            logger.info(f'found a match on OpenAI with {sanitized_openai_text}=={sanitized_expected_text}')
            return
        else:
            logger.warning(f'failed to recognize text using OpenAI. expected text: {sanitized_expected_text} got: {sanitized_openai_text}')

        # second, try azure
        logger.info(f'attempting to recognize text using Azure. expected text: {sanitized_expected_text}')
        audio_text = audio_utils.speech_to_text_azure_wav(self.manager, audio_temp_file, recognition_language, audio_format)
        sanitized_azure_text = audio_utils.sanitize_recognized_text(audio_text)
        self.assertEqual(sanitized_expected_text, sanitized_azure_text)

    def verify_service_audio_language(self, text, service, audio_language, recognition_language):
        """Legacy version using voice_list"""
        # logging.info(f'verify_service_audio: service: {service} audio_language: {audio_language}')
        voices = self.get_voice_list_service_audio_language(service, audio_language)
        self.assertGreaterEqual(len(voices), 1, f'at least one voice for service {service}, language {audio_language}')

        if service == Service.Google:
            # exclude Journey voices, they don't use the standard interface
            voices = [x for x in voices if 'Journey' not in x['voice_name']]

        # pick 3 random voices
        max_voices = 3
        if len(voices) > max_voices:
            voices = random.sample(voices, max_voices)
        for voice in voices:
            self.verify_voice(voice, text, recognition_language)

    def verify_service_japanese(self, service: Service):
        source_text = self.JAPANESE_INPUT_TEXT
        self.verify_service_audio_language(source_text, service, AudioLanguage.ja_JP, 'ja-JP')

    def verify_service_chinese(self, service: Service):
        source_text = self.CHINESE_INPUT_TEXT
        self.verify_service_audio_language(source_text, service, AudioLanguage.zh_CN, 'zh-CN')

    def verify_service_korean(self, service: Service):
        source_text = self.KOREAN_INPUT_TEXT
        self.verify_service_audio_language(source_text, service, AudioLanguage.ko_KR, 'ko-KR')

    def test_voice_list_v3(self):
        # pytest tests/test_audio.py -k test_voice_list_v3
        voice_list_v3 = self.voice_list_v3
        # there should be more than 1000 voices in total
        self.assertTrue(len(voice_list_v3) > 1000)

        # check Azure voices
        # ==================

        # search for some particular voices
        azure_voices = [x for x in voice_list_v3 if x.service == Service.Azure]
        mandarin_azure_voices = [x for x in azure_voices if AudioLanguage.zh_CN in x.audio_languages]
        self.assertTrue(len(mandarin_azure_voices) > 10)

        # pprint.pprint(mandarin_azure_voices)

        xiaochen = [x for x in mandarin_azure_voices if 'Xiaochen' in x.name]
        logger.debug(f'xiaochen voices: {pprint.pformat(xiaochen)}')
        self.assertIn(len(xiaochen), [3, 4], str(xiaochen)) # there is a regular and a multilingual
        # and also a DragonHD, and DragonHD V1

        xiaochen_single_language = [x for x in xiaochen if len(x.audio_languages) == 1][0]
        self.assertEqual(xiaochen_single_language.audio_languages, [AudioLanguage.zh_CN])

        xiaochen_multilingual = [x for x in xiaochen if len(x.audio_languages) > 1][0]
        self.assertTrue(len(xiaochen_multilingual.audio_languages) > 1)
        self.assertTrue(AudioLanguage.zh_CN in xiaochen_multilingual.audio_languages)
        self.assertTrue(AudioLanguage.fr_FR in xiaochen_multilingual.audio_languages)

        # check OpenAI voices
        # ===================
        openai_voices = [x for x in voice_list_v3 if x.service == Service.OpenAI]
        self.assertTrue(len(openai_voices) >= 6)
        # choose random voice
        openai_voice = random.choice(openai_voices)
        self.assertTrue(len(openai_voice.audio_languages) >= 57) # at least 57 locales supportedn

        # check ElevenLabs voices
        # =======================
        elevenlabs_voices = [x for x in voice_list_v3 if x.service == Service.ElevenLabs]
        self.assertTrue(len(elevenlabs_voices) >= 15)
        mandarin_elevenlabs_voices = [x for x in elevenlabs_voices if AudioLanguage.zh_CN in x.audio_languages]
        self.assertTrue(len(mandarin_elevenlabs_voices) > 10)


    def test_french_google(self):
        source_text = self.FRENCH_INPUT_TEXT
        self.verify_service_audio_language(source_text, Service.Google, AudioLanguage.fr_FR, 'fr-FR')

    def test_french_azure(self):
        source_text = self.FRENCH_INPUT_TEXT
        self.verify_service_audio_language(source_text, Service.Azure, AudioLanguage.fr_FR, 'fr-FR')

    def test_french_watson(self):
        source_text = self.FRENCH_INPUT_TEXT
        self.verify_service_audio_language(source_text, Service.Watson, AudioLanguage.fr_FR, 'fr-FR')

    def test_french_vocalware(self):
        source_text = self.FRENCH_INPUT_TEXT
        self.verify_service_audio_language(source_text, Service.VocalWare, AudioLanguage.fr_FR, 'fr-FR')        

    def test_english_amazon(self):
        source_text = 'I am not interested.'
        self.verify_service_audio_language(source_text, Service.Amazon, AudioLanguage.en_GB, 'en-GB')

    def test_french_amazon(self):
        source_text = self.FRENCH_INPUT_TEXT
        self.verify_service_audio_language(source_text, Service.Amazon, AudioLanguage.fr_FR, 'fr-FR')        

    def test_mandarin_google(self):
        source_text = self.CHINESE_INPUT_TEXT
        self.verify_service_audio_language(source_text, Service.Google, AudioLanguage.zh_CN, 'zh-CN')

    def test_mandarin_azure(self):
        source_text = self.CHINESE_INPUT_TEXT
        self.verify_service_audio_language(source_text, Service.Azure, AudioLanguage.zh_CN, 'zh-CN') 

    def test_azure_standard_voice_deprecated(self):
        # standard voices should raise an error
        service = "Azure"
        source_text = "流暢"
        voice_key= {
            "name": "Microsoft Server Speech Text to Speech Voice (ja-JP, HarukaRUS)"
        }
        options = {} 

        with self.assertRaises(cloudlanguagetools.errors.RequestError) as cm:
            self.manager.get_tts_audio(source_text, service, voice_key, options)

        self.assertEqual(str(cm.exception), 'Azure Standard voices are not supported anymore, please switch to Neural voices.')

    def test_mandarin_azure_xiaochen_multilingual(self):
        # search for some particular voices
        azure_voices = [x for x in self.voice_list_v3 if x.service == Service.Azure]
        mandarin_azure_voices = [x for x in azure_voices if AudioLanguage.zh_CN in x.audio_languages]

        xiaochen = [x for x in mandarin_azure_voices if 'Xiaochen' in x.name]
        self.assertIn(len(xiaochen), [3, 4]) # there is a regular and a multilingual, and dragonhd, 
        # and dragonhd latest

        xiaochen_single_language = [x for x in xiaochen if len(x.audio_languages) == 1][0]
        xiaochen_multilingual = [x for x in xiaochen if len(x.audio_languages) > 1][0]

        # check that single language voice can generate chinese
        self.verify_voice_v3(xiaochen_single_language, self.CHINESE_INPUT_TEXT, 'zh-CN')

        # check that multilingual voice can generate chinese and french
        self.verify_voice_v3(xiaochen_multilingual, self.CHINESE_INPUT_TEXT, 'zh-CN')
        self.verify_voice_v3(xiaochen_multilingual, self.FRENCH_INPUT_TEXT, 'fr-FR')

    def test_openai_multilingual_v3(self):
        # get random voice
        openai_voices = [x for x in self.voice_list_v3 if x.service == Service.OpenAI]
        selected_voice = random.choice(openai_voices)

        source_text = 'This is the best restaurant in town.'
        self.verify_voice_v3(selected_voice, source_text, 'en-US')

        source_text = self.CHINESE_INPUT_TEXT
        self.verify_voice_v3(selected_voice, source_text, 'zh-CN')

    def test_elevenlabs_multilingual_v3(self):
        # get random voice
        elevenlabs_voices = [x for x in self.voice_list_v3 if x.service == Service.ElevenLabs]
        selected_voice = random.choice(elevenlabs_voices)

        source_text = 'This is the best restaurant in town.'
        self.verify_voice_v3(selected_voice, source_text, 'en-US')

        # chinese seems to be unreliable
        # source_text = self.CHINESE_INPUT_TEXT
        # self.verify_voice_v3(selected_voice, source_text, 'zh-CN')

        source_text = self.FRENCH_INPUT_TEXT
        self.verify_voice_v3(selected_voice, source_text, 'fr-FR')


    @pytest.mark.skip('watson does not support chinese anymore')
    def test_mandarin_watson(self):
        # pytest tests/test_audio.py -s --log-cli-level=DEBUG -k test_mandarin_watson
        source_text = '老人家'
        self.verify_service_audio_language(source_text, Service.Watson, AudioLanguage.zh_CN, 'zh-CN')

    def test_mandarin_amazon(self):
        source_text = '老人家'
        self.verify_service_audio_language(source_text, Service.Amazon, AudioLanguage.zh_CN, 'zh-CN')

    def test_mandarin_cereproc(self):
        # pytest test_audio.py -k test_mandarin_cereproc
        source_text = '老人家'
        self.verify_service_audio_language(source_text, Service.CereProc, AudioLanguage.zh_CN, 'zh-CN')

    def test_mandarin_alibaba(self):
        # pytest test_audio.py -k test_mandarin_alibaba
        source_text = '老人家'
        self.verify_service_audio_language_v3(source_text, Service.Alibaba, AudioLanguage.zh_CN, 'zh-CN')

    def test_english_alibaba(self):
        # pytest test_audio.py -k test_english_alibaba
        source_text = 'I am not interested.'
        self.verify_service_audio_language_v3(source_text, Service.Alibaba, AudioLanguage.en_GB, 'en-GB')

    @skip_unreliable_clt_test()
    def test_mandarin_vocalware(self):
        # pytest test_audio.py -k test_mandarin_vocalware
        
        source_text = '你好'
        self.verify_service_audio_language(source_text, Service.VocalWare, AudioLanguage.zh_CN, 'zh-CN')

    def test_cantonese_google(self):
        source_text = self.CHINESE_INPUT_TEXT
        self.verify_service_audio_language(source_text, Service.Google, AudioLanguage.zh_HK, 'zh-HK')

    def test_cantonese_azure(self):
        source_text = '天氣預報'
        self.verify_service_audio_language(source_text, Service.Azure, AudioLanguage.zh_HK, 'zh-HK')

    def test_cantonese_simplified_azure(self):
        source_text = '天氣預報'
        self.verify_service_audio_language(source_text, Service.Azure, AudioLanguage.yue_CN, 'zh-HK')        

    def test_korean_naver(self):
        # pytest test_audio.py -k test_korean_naver
        source_text = self.KOREAN_INPUT_TEXT
        self.verify_service_audio_language(source_text, Service.Naver, AudioLanguage.ko_KR, 'ko-KR')

    def test_japanese_naver(self):
        # pytest test_audio.py -k test_japanese_naver
        source_text = self.JAPANESE_INPUT_TEXT
        self.verify_service_audio_language(source_text, Service.Naver, AudioLanguage.ja_JP, 'ja-JP')

    def test_japanese_naver_error(self):
        # pytest test_audio.py -k test_japanese_naver_error

        source_text = '瑩'
        service = 'Naver'
        voice_key = {"name": "ntomoko"}
        options = {}

        self.assertRaises(cloudlanguagetools.errors.RequestError, 
                          self.manager.get_tts_audio,
                          source_text, service, voice_key, options )


    def test_vietnamese_fptai(self):
        # pytest test_audio.py -k test_vietnamese_fptai
        source_text = 'thì giá là bao nhiêu'
        self.verify_service_audio_language(source_text, Service.FptAi, AudioLanguage.vi_VN, 'vi-VN')

    def test_english_naver(self):
        # pytest test_audio.py -k test_english_naver
        source_text = 'this is the first sentence'
        # source_text = 'hello'
        self.verify_service_audio_language(source_text, Service.Naver, AudioLanguage.en_US, 'en-US')

    def test_english_vocalware(self):
        # pytest test_audio.py -k test_english_vocalware
        source_text = 'this is the first sentence'
        # source_text = 'hello'
        self.verify_service_audio_language(source_text, Service.VocalWare, AudioLanguage.en_US, 'en-US')


    def test_ssml_english_google(self):
        # pytest tests/test_audio.py -k test_ssml_english_google
        if not CLOUDLANGUAGETOOLS_CORE_TEST_UNRELIABLE:
            pytest.skip('you must set CLOUDLANGUAGETOOLS_CORE_TEST_UNRELIABLE=yes')

        source_text = 'hello <break time="50ms"/>world'
        self.verify_service_audio_language(source_text, Service.Google, AudioLanguage.en_US, 'en-US')

    @skip_unreliable_clt_test()
    def test_ssml_english_azure(self):
        # pytest tests/test_audio.py -k test_ssml_english_azure
        source_text = 'hello <break time="200ms"/>world'
        self.verify_service_audio_language(source_text, Service.Azure, AudioLanguage.en_US, 'en-US')

    def test_ssml_english_amazon(self):
        # pytest test_audio.py -k test_ssml_english_amazon

        us_standard_neural_voices = [x for x in self.voice_list_v3 if 
                  x.service == Service.Amazon and
                  x.voice_key['engine'] in ['standard', 'neural'] and 
            AudioLanguage.en_US in x.audio_languages]

        # choose random voice from standard_neural_voices
        standard_neural_voice = random.choice(us_standard_neural_voices)
        source_text = 'hello <break time="200ms"/>world'
        self.verify_voice_v3(standard_neural_voice, source_text, 'en-US')


    def test_ssml_english_watson(self):
        # pytest test_audio.py -k test_ssml_english_watson
        source_text = 'hello <break time="200ms"/>world'
        self.verify_service_audio_language(source_text, Service.Watson, AudioLanguage.en_US, 'en-US')

    def test_ssml_english_cereproc(self):
        # pytest test_audio.py -k test_ssml_english_cereproc
        if not CLOUDLANGUAGETOOLS_CORE_TEST_UNRELIABLE:
            pytest.skip('you must set CLOUDLANGUAGETOOLS_CORE_TEST_UNRELIABLE=yes')

        source_text = 'hello <break time="200ms"/>my friends'
        self.verify_service_audio_language(source_text, Service.CereProc, AudioLanguage.en_GB, 'en-GB')

    def test_english_forvo(self):
        # pytest test_audio.py -k test_english_forvo
        source_text = 'absolutely'
        self.verify_service_audio_language(source_text, Service.Forvo, AudioLanguage.en_US, 'en-US')

    @pytest.mark.skip('not stable')
    def test_english_cereproc(self):
        # pytest test_audio.py -rPP -k test_english_cereproc
        source_text = 'absolutely'
        self.verify_service_audio_language(source_text, Service.CereProc, AudioLanguage.en_GB, 'en-GB')

    def test_english_forvo_preferred_user(self):
        # pytest test_audio.py -rPP -k test_english_forvo_preferred_user
        source_text = 'vehicle'
        service = 'Forvo'
        voice_key = {
            "country_code": "ANY",
            "language_code": "en",
            'preferred_user': 'lizchiu'
        }        
        options = {}
        audio_temp_file = self.manager.get_tts_audio(source_text, service, voice_key, options)
        audio_text = audio_utils.speech_to_text(self.manager, audio_temp_file, 'en-GB')
        self.assertEqual(audio_utils.sanitize_recognized_text(source_text), audio_utils.sanitize_recognized_text(audio_text))

    def test_french_forvo(self):
        # pytest test_audio.py -k test_french_forvo
        source_text = 'absolument'
        self.verify_service_audio_language(source_text, Service.Forvo, AudioLanguage.fr_FR, 'fr-FR')

    def test_portugal_portuguese_forvo(self):
        # pytest test_audio.py -k test_brazilian_portuguese_forvo
        source_text = 'obrigado'
        self.verify_service_audio_language(source_text, Service.Forvo, AudioLanguage.pt_PT, 'pt-PT')

    def test_brazilian_portuguese_forvo(self):
        # pytest test_audio.py -k test_brazilian_portuguese_forvo
        source_text = 'obrigado'
        self.verify_service_audio_language(source_text, Service.Forvo, AudioLanguage.pt_BR, 'pt-BR')

    def test_mandarin_forvo(self):
        # pytest tests/test_audio.py -k test_mandarin_forvo
        source_text = '你好'
        self.verify_service_audio_language(source_text, Service.Forvo, AudioLanguage.zh_CN, 'zh-CN')

    def test_french_cereproc(self):
        # pytest test_audio.py -k test_french_cereproc
        source_text = 'absolument'
        self.verify_service_audio_language(source_text, Service.CereProc, AudioLanguage.fr_FR, 'fr-FR')

    def test_french_forvo_not_found(self):
        # pytest test_audio.py -k test_french_forvo_not_found
        source_text = 'wordnotfound'
        self.assertRaises(cloudlanguagetools.errors.NotFoundError, self.verify_service_audio_language, source_text, Service.Forvo, AudioLanguage.fr_FR, 'fr-FR')

    @pytest.mark.skip(reason="voicen service has disappeared")
    def test_russian_voicen(self):
        # pytest test_audio.py -k test_russian_voicen
        source_text = 'улица'
        self.verify_service_audio_language(source_text, Service.Voicen, AudioLanguage.ru_RU, 'ru-RU')

    @pytest.mark.skip(reason="voicen service has disappeared")
    def test_turkish_voicen(self):
        # pytest test_audio.py -k test_turkish_voicen
        source_text = 'kahvaltı'
        self.verify_service_audio_language(source_text, Service.Voicen, AudioLanguage.tr_TR, 'tr-TR')

    def test_azure_options(self):
        service = 'Azure'
        source_text = self.FRENCH_INPUT_TEXT

        voice_key = {
            "name": "Microsoft Server Speech Text to Speech Voice (fr-FR, DeniseNeural)"
        }

        options = {'rate': 0.8, 'pitch': -10}

        audio_temp_file = self.manager.get_tts_audio(source_text, service, voice_key, options)

        self.recognize_and_verify_text(
            audio_temp_file, source_text, 'fr-FR', cloudlanguagetools.options.AudioFormat.mp3)

    def test_azure_format_ogg(self):
        service = 'Azure'
        source_text = self.FRENCH_INPUT_TEXT

        voice_key = {
            "name": "Microsoft Server Speech Text to Speech Voice (fr-FR, DeniseNeural)"
        }

        options = {'rate': 0.8, 'pitch': -10, 'format': 'ogg_opus'}

        audio_temp_file = self.manager.get_tts_audio(source_text, service, voice_key, options)

        self.assertTrue(audio_utils.is_ogg_opus_format(audio_temp_file.name))
        
        audio_text = audio_utils.speech_to_text(self.manager, audio_temp_file, 'fr-FR', audio_format=cloudlanguagetools.options.AudioFormat.ogg_opus)
        self.assertEqual(audio_utils.sanitize_recognized_text(source_text), audio_utils.sanitize_recognized_text(audio_text))        

    def verify_service_audio_language_v3(self, text, service, audio_language, recognition_language):
        """Version using voice_list_v3"""
        voices = [x for x in self.voice_list_v3 if x.service == service and audio_language in x.audio_languages]
        self.assertGreaterEqual(len(voices), 1, f'at least one voice for service {service}, language {audio_language}')

        # pick 3 random voices
        max_voices = 3
        if len(voices) > max_voices:
            voices = random.sample(voices, max_voices)
        for voice in voices:
            self.verify_voice_v3(voice, text, recognition_language)

    def verify_wav_voice(self, voice: cloudlanguagetools.ttsvoice.TtsVoice_v3, text: str, recognition_language: str):
        # assert that the wav format is in the list of supported formats
        self.assertTrue(cloudlanguagetools.options.AudioFormat.wav.name in 
                        voice.options[cloudlanguagetools.options.AUDIO_FORMAT_PARAMETER]['values'])
        options = {cloudlanguagetools.options.AUDIO_FORMAT_PARAMETER: cloudlanguagetools.options.AudioFormat.wav.name}
        audio_temp_file = self.manager.get_tts_audio(text, voice.service, voice.voice_key, options)
        audio_utils.assert_is_wav_format(self, audio_temp_file.name)

        self.recognize_and_verify_text(
            audio_temp_file, 
            text, 
            recognition_language, 
            cloudlanguagetools.options.AudioFormat.wav)

    def test_azure_format_wav(self):
        fr_voice = self.get_voice_by_service_and_name(Service.Azure, 'Denise')
        self.verify_wav_voice(fr_voice, self.FRENCH_INPUT_TEXT, 'fr-FR')

    def test_amazon_format_wav(self):
        fr_voice = self.get_voice_by_service_and_name(Service.Amazon, 'Mathieu')
        self.verify_wav_voice(fr_voice, self.FRENCH_INPUT_TEXT, 'fr-FR')

    def test_amazon_generative(self):
        amy_generative_voice = self.get_voice_by_lambda(Service.Amazon, 
            lambda x: x.voice_key['engine'] == 'generative' and 'Amy' in x.name)
        self.verify_voice_v3(amy_generative_voice, self.ENGLISH_INPUT_TEXT, 'en-GB')

    def test_elevenlabs_format_wav(self):
        fr_voice = self.get_voice_by_lambda(Service.ElevenLabs, 
            lambda x: 'Charlotte' in x.name and x.voice_key['model_id'] == 'eleven_multilingual_v2')
        self.verify_wav_voice(fr_voice, self.FRENCH_INPUT_TEXT, 'fr-FR')        

    def test_google_format_wav(self):
        fr_voice = self.get_voice_by_lambda(Service.Google, 
            lambda x: AudioLanguage.fr_FR in x.audio_languages and 'Journey' not in x.name, assert_unique=False)
        self.verify_wav_voice(fr_voice, self.FRENCH_INPUT_TEXT, 'fr-FR')

    def test_naver_format_wav(self):
        ko_voice = self.get_voice_by_lambda(Service.Naver,
            lambda x: AudioLanguage.ko_KR in x.audio_languages, assert_unique=False)
        self.verify_wav_voice(ko_voice, self.KOREAN_INPUT_TEXT, 'ko-KR')

    def test_openai_format_wav(self):
        ko_voice = self.get_voice_by_lambda(Service.OpenAI,
            lambda x: AudioLanguage.en_US in x.audio_languages, assert_unique=False)
        self.verify_wav_voice(ko_voice, self.ENGLISH_INPUT_TEXT, 'en-US')

    def test_watson_format_wav(self):
        en_voice = self.get_voice_by_lambda(Service.Watson,
            lambda x: AudioLanguage.en_US in x.audio_languages, assert_unique=False)
        self.verify_wav_voice(en_voice, self.ENGLISH_INPUT_TEXT, 'en-US')


    @pytest.mark.skip(reason="journey voice seems to be gone")
    def test_google_voice_journey_old(self):
        service = 'Google'
        source_text = self.ENGLISH_INPUT_TEXT

        voice_key = {
            'name': 'en-US-Journey-F', 
            'language_code': 'en-US', 
            'ssml_gender': 'FEMALE'}

        # seems to return:
        # google.api_core.exceptions.InvalidArgument: 400 This voice currently only supports LINEAR16 and MULAW output.
        # ensure that exception is properly wrapped
        self.assertRaises(cloudlanguagetools.errors.RequestError, self.manager.get_tts_audio, source_text, service, voice_key, {})

    def test_google_format_ogg(self):
        service = 'Google'
        source_text = self.FRENCH_INPUT_TEXT

        voice_key = {
            'language_code': 'fr-FR', 
            'name': 'fr-FR-Wavenet-E', 
            'ssml_gender': 'FEMALE'
        }

        options = {'format': 'ogg_opus'}

        audio_temp_file = self.manager.get_tts_audio(source_text, service, voice_key, options)

        file_type = magic.from_file(audio_temp_file.name)
        self.assertIn('Ogg data, Opus', file_type)
        
        audio_text = audio_utils.speech_to_text(self.manager, audio_temp_file, 'fr-FR', audio_format=cloudlanguagetools.options.AudioFormat.ogg_opus)
        self.assertEqual(audio_utils.sanitize_recognized_text(source_text), audio_utils.sanitize_recognized_text(audio_text))        

    def test_amazon_format_ogg(self):
        service = 'Amazon'
        source_text = self.FRENCH_INPUT_TEXT

        voice_key = {'engine': 'neural', 'voice_id': 'Gabrielle'}

        options = {'format': 'ogg_vorbis'}

        audio_temp_file = self.manager.get_tts_audio(source_text, service, voice_key, options)

        file_type = magic.from_file(audio_temp_file.name)
        self.assertIn('Ogg data', file_type)
        
        audio_text = audio_utils.speech_to_text(self.manager, audio_temp_file, 'fr-FR', audio_format=cloudlanguagetools.options.AudioFormat.ogg_vorbis)
        self.assertEqual(audio_utils.sanitize_recognized_text(source_text), audio_utils.sanitize_recognized_text(audio_text))        

    def test_azure_ampersand(self):
        service = 'Azure'
        source_text = 'In my ignorance I had never heard country & western music.'

        voice_key = {
            "name": "Microsoft Server Speech Text to Speech Voice (en-US, AriaNeural)"
        }

        audio_temp_file = self.manager.get_tts_audio(source_text, service, voice_key, {})
        audio_text = audio_utils.speech_to_text(self.manager, audio_temp_file, 'en-US')
        self.assertEqual(audio_utils.sanitize_recognized_text(source_text), audio_utils.sanitize_recognized_text(audio_text))        

    def test_fptai_options(self):
        service = 'FptAi'
        source_text = 'xin chào bạn của tôi'

        voice_key = {
            "voice_id": "leminh"
        }

        options = {'speed': -0.5}

        audio_temp_file = self.manager.get_tts_audio(source_text, service, voice_key, options)
        audio_text = audio_utils.speech_to_text(self.manager, audio_temp_file, 'vi-VN')
        self.assertEqual(audio_utils.sanitize_recognized_text(source_text), audio_utils.sanitize_recognized_text(audio_text))

    def test_elevenlabs_english(self):
        source_text = 'This is the best restaurant in town.'
        self.verify_service_audio_language(source_text, Service.ElevenLabs, AudioLanguage.en_US, 'en-US')

    def test_elevenlabs_english_all_voices_charlotte(self):
        # pytest tests/test_audio.py -s --log-cli-level=INFO -k test_elevenlabs_english_all_voices_charlotte
        source_text = 'This is the best restaurant in town.'
        voice_list = self.get_voice_list_service_audio_language(Service.ElevenLabs, AudioLanguage.en_US)
        # pprint.pprint(voice_list)
        selected_voice_candidates = [x for x in voice_list if x['audio_language_code'] == 'en_US' and x['service'] == 'ElevenLabs']
        selected_voice_candidates = [x for x in selected_voice_candidates if 'Charlotte' in x['voice_name']]
        pprint.pprint(selected_voice_candidates)
        logger.info(f'number of voices: {len(selected_voice_candidates)}')
        # self.assertTrue(False)
        for voice in selected_voice_candidates:
            voice_str = f"{voice['voice_name']} {voice['voice_key']}"
            logger.info(f"testing voice {voice_str}")
            self.verify_voice(voice, source_text, 'en-US')
            logger.info(f'voice: {voice_str} passed')

    def test_elevenlabs_french(self):
        source_text = self.FRENCH_INPUT_TEXT
        self.verify_service_audio_language(source_text, Service.ElevenLabs, AudioLanguage.fr_FR, 'fr-FR')


    def test_elevenlabs_spanish(self):
        source_text = 'No me gusta la comida.'
        self.verify_service_audio_language(source_text, Service.ElevenLabs, AudioLanguage.es_ES, 'es-ES')

    def test_elevenlabs_italian(self):
        source_text = 'Non mi piace il cibo.'
        self.verify_service_audio_language(source_text, Service.ElevenLabs, AudioLanguage.it_IT, 'it-IT')

    def test_elevenlabs_german(self):
        source_text = 'Ich mag das Essen nicht.'
        self.verify_service_audio_language(source_text, Service.ElevenLabs, AudioLanguage.de_DE, 'de-DE')

    def test_elevenlabs_german(self):
        source_text = 'Ich mag das Essen nicht.'
        self.verify_service_audio_language(source_text, Service.ElevenLabs, AudioLanguage.de_DE, 'de-DE')        

    @skip_unreliable_clt_test()
    def test_elevenlabs_japanese(self):
        self.verify_service_japanese(Service.ElevenLabs)

    @skip_unreliable_clt_test()
    def test_elevenlabs_chinese(self):
        self.verify_service_chinese(Service.ElevenLabs)

    @skip_unreliable_clt_test()
    def test_elevenlabs_korean(self):
        self.verify_service_korean(Service.ElevenLabs)

    def test_openai_english(self):
        source_text = 'This is the best restaurant in town.'
        self.verify_service_audio_language(source_text, Service.OpenAI, AudioLanguage.en_US, 'en-US')

    @pytest.mark.skip(reason="detection of openai french is unreliable")
    def test_openai_french(self):
        source_text = self.FRENCH_INPUT_TEXT
        self.verify_service_audio_language(source_text, Service.OpenAI, AudioLanguage.fr_FR, 'fr-FR')



import json
import requests
import tempfile
import uuid
import operator
import pydub
import logging
import pprint
import time
import cachetools
from typing import List

import cloudlanguagetools.service
import cloudlanguagetools.constants
import cloudlanguagetools.options
import cloudlanguagetools.languages
import cloudlanguagetools.ttsvoice
import cloudlanguagetools.translationlanguage
import cloudlanguagetools.transliterationlanguage
import cloudlanguagetools.dictionarylookup
import cloudlanguagetools.errors
from cloudlanguagetools.options import AudioFormat


import azure.cognitiveservices.speech
import azure.cognitiveservices.speech.audio

logger = logging.getLogger(__name__)

AUDIO_LOCALE_OVERRIDE_MAP = {
    'sr-Latn-RS': 'sr_RS'
}

GENDER_MAP = {
    'Female': cloudlanguagetools.constants.Gender.Female,
    'Male': cloudlanguagetools.constants.Gender.Male,
    'Neutral': cloudlanguagetools.constants.Gender.Any,
}

VOICE_OPTIONS = {
            'rate' : {
                'type': cloudlanguagetools.options.ParameterType.number.name,
                'min': 0.5,
                'max': 3.0,
                'default': 1.0
            },
            'pitch': {
                'type': cloudlanguagetools.options.ParameterType.number.name,
                'min': -100,
                'max': 100,
                'default': 0
            },
            cloudlanguagetools.options.AUDIO_FORMAT_PARAMETER: {
                'type': cloudlanguagetools.options.ParameterType.list.name,
                'values': [
                    cloudlanguagetools.options.AudioFormat.mp3.name,
                    cloudlanguagetools.options.AudioFormat.ogg_opus.name,
                    cloudlanguagetools.options.AudioFormat.wav.name,
                ],
                'default': cloudlanguagetools.options.AudioFormat.mp3.name
            }
}

class AzureVoice(cloudlanguagetools.ttsvoice.TtsVoice):
    def __init__(self, voice_data):
        # print(voice_data)
        self.service = cloudlanguagetools.constants.Service.Azure
        self.service_fee = cloudlanguagetools.constants.ServiceFee.paid
        locale = voice_data['Locale']
        locale = AUDIO_LOCALE_OVERRIDE_MAP.get(locale, locale)
        language_enum_name = locale.replace('-', '_')
        self.audio_language = cloudlanguagetools.languages.AudioLanguage[language_enum_name]
        self.name = voice_data['Name']
        self.display_name = voice_data['DisplayName']
        self.local_name = voice_data['LocalName']
        self.short_name = voice_data['ShortName']
        self.gender = GENDER_MAP[voice_data['Gender']]

        self.locale = locale
        self.voice_type = voice_data['VoiceType']


    def get_voice_key(self):
        return {
            'name': self.name
        }

    def get_voice_shortname(self):
        if self.local_name != self.display_name:
            return f'{self.display_name} {self.local_name} ({self.voice_type})'
        else:
            return f'{self.display_name} ({self.voice_type})'

    def get_options(self):
        return VOICE_OPTIONS

def locale_to_audio_language(locale: str) -> cloudlanguagetools.languages.AudioLanguage:

    locale_override_map = {
        'zh-CN-shanxi': 'zh-CN-shaanxi'
    }
    locale = locale_override_map.get(locale, locale)

    locale = AUDIO_LOCALE_OVERRIDE_MAP.get(locale, locale)
    language_enum_name = locale.replace('-', '_')
    audio_language = cloudlanguagetools.languages.AudioLanguage[language_enum_name]    
    return audio_language

def build_tts_voice_v3(voice_data) -> cloudlanguagetools.ttsvoice.TtsVoice_v3:
    local_name = voice_data['LocalName']
    display_name = voice_data['DisplayName']
    voice_type = voice_data['VoiceType']

    # build all attributes required for TtsVoice_v3
    # name
    if local_name != display_name:
        voice_name = f"{display_name} {local_name} ({voice_type})"
    else:
        voice_name = f'{display_name} ({voice_type})'    
    voice_key = {
        'name': voice_data['Name']
    }
    options = VOICE_OPTIONS
    service = cloudlanguagetools.constants.Service.Azure
    gender = GENDER_MAP[voice_data['Gender']]
    service_fee = cloudlanguagetools.constants.ServiceFee.paid

    azure_locale_list = [voice_data['Locale']]
    if 'SecondaryLocaleList' in voice_data:
        azure_locale_list = voice_data['SecondaryLocaleList']
        # ensure the main locale is present
        azure_locale_list.append(voice_data['Locale'])
        # unique array
        azure_locale_list = list(set(azure_locale_list))

    audio_languages = [locale_to_audio_language(locale) for locale in azure_locale_list]

    return cloudlanguagetools.ttsvoice.TtsVoice_v3(
        name=voice_name,
        voice_key=voice_key,
        options=options,
        service=service,
        gender=gender,
        audio_languages=audio_languages,
        service_fee=service_fee)


def get_translation_language_enum(language_id):
    # print(f'language_id: {language_id}')
    azure_language_id_map = {
        'as': 'as_',
        'fr-ca': 'fr_ca',
        'fr-CA': 'fr_ca',
        'id': 'id_',
        'is': 'is_',
        'or': 'or_',
        'pt': 'pt_br',
        'pt-pt': 'pt_pt',
        'pt-PT': 'pt_pt',
        'sr-Cyrl': 'sr_cyrl',
        'sr-Latn': 'sr_latn',
        'tlh-Latn': 'tlh_latn',
        'tlh-Piqd': 'tlh_piqd',
        'zh-Hans': 'zh_cn',
        'zh-Hant': 'zh_tw',
        'lzh': 'zh_lit',
        'mn-Cyrl': 'mn_cyrl',
        'mn-Mong': 'mn',
        'iu-Latn': 'iu_latn',
        'lug': 'lg',
        'nya': 'ny',
        'run': 'rn',
    }
    if language_id in azure_language_id_map:
        language_id = azure_language_id_map[language_id]
    return cloudlanguagetools.languages.Language[language_id]

class AzureTranslationLanguage(cloudlanguagetools.translationlanguage.TranslationLanguage):
    def __init__(self, language_id):
        self.service = cloudlanguagetools.constants.Service.Azure
        self.service_fee = cloudlanguagetools.constants.ServiceFee.paid
        self.language_id = language_id
        self.language = get_translation_language_enum(language_id)

    def get_language_id(self):
        return self.language_id

class AzureTransliterationLanguage(cloudlanguagetools.transliterationlanguage.TransliterationLanguage):
    def __init__(self, language_id, from_script, to_script, from_script_name, to_script_name, from_native_name, to_native_name):
        self.service = cloudlanguagetools.constants.Service.Azure
        self.service_fee = cloudlanguagetools.constants.ServiceFee.paid
        self.language_id = language_id
        self.language = get_translation_language_enum(language_id)
        self.from_script = from_script
        self.to_script = to_script
        self.from_script_name = from_script_name
        self.to_script_name = to_script_name
        self.from_native_name = from_native_name
        self.to_native_name = to_native_name

    def get_transliteration_name(self):
        result = f'{self.language.lang_name} ({self.from_script_name}/{self.from_native_name} to {self.to_script_name}/{self.to_native_name}), {self.service.name}'
        return result

    def get_transliteration_shortname(self):
        result = f'{self.from_script_name}/{self.from_native_name} to {self.to_script_name}/{self.to_native_name}, {self.service.name}'
        return result        

    def get_transliteration_key(self):
        return {
            'language_id': self.language_id,
            'from_script': self.from_script,
            'to_script': self.to_script
        }

class AzureDictionaryLookup(cloudlanguagetools.dictionarylookup.DictionaryLookup):
    def __init__(self, source_language_code, target_language_code, lookup_type):
        self.service = cloudlanguagetools.constants.Service.Azure
        self.service_fee = cloudlanguagetools.constants.ServiceFee.paid
        self.language = get_translation_language_enum(source_language_code)
        self.target_language = get_translation_language_enum(target_language_code)
        self.source_language_code = source_language_code
        self.target_language_code = target_language_code
        self.lookup_type = lookup_type

    def get_lookup_key(self):
        return {
            'source_language_code': self.source_language_code,
            'target_language_code': self.target_language_code,
            'language': self.language.name,
            'lookup_type': self.lookup_type.name
        }        

    def get_lookup_name(self):
        return f'{self.service.name} ({self.language.lang_name} to {self.target_language.lang_name}), {self.lookup_type.name}'

    def get_lookup_shortname(self):
        return f'{self.service.name}, {self.language.lang_name}, {self.lookup_type.name}'

class AzureService(cloudlanguagetools.service.Service):
    def __init__(self):
        self.url_translator_base = 'https://api.cognitive.microsofttranslator.com'

    def configure(self, config):
        self.key = config['key']
        self.region = config['region']

    def get_token(self):
        fetch_token_url = f"https://{self.region}.api.cognitive.microsoft.com/sts/v1.0/issueToken"
        headers = {
            'Ocp-Apim-Subscription-Key': self.key
        }
        response = requests.post(fetch_token_url, headers=headers, timeout=cloudlanguagetools.constants.RequestTimeout)
        access_token = str(response.text)
        return access_token

    def get_translator_headers(self):
        headers = {
            'Ocp-Apim-Subscription-Key': self.key,
            'Ocp-Apim-Subscription-Region': self.region,
            'Content-type': 'application/json',
            'X-ClientTraceId': str(uuid.uuid4())
        }
        return headers        

    def get_tts_audio(self, text, voice_key, options):
        # https://learn.microsoft.com/en-us/azure/ai-services/speech-service/rest-text-to-speech?tabs=streaming#audio-outputs
        # https://learn.microsoft.com/en-us/python/api/azure-cognitiveservices-speech/azure.cognitiveservices.speech.speechsynthesisoutputformat?view=azure-python
        response_format_parameter, audio_format = self.get_request_audio_format({
            AudioFormat.mp3: 'Audio24Khz96KBitRateMonoMp3',
            AudioFormat.ogg_opus: 'Ogg48Khz16BitMonoOpus',
            AudioFormat.wav: 'Riff48Khz16BitMonoPcm'
        }, options, AudioFormat.mp3)

        output_temp_file = tempfile.NamedTemporaryFile(prefix=f'cloudlanguage_tools_{self.__class__.__name__}_audio', suffix=f'.{audio_format.name}')
        output_temp_filename = output_temp_file.name
        speech_config = azure.cognitiveservices.speech.SpeechConfig(subscription=self.key, region=self.region)
        speech_config.set_speech_synthesis_output_format(azure.cognitiveservices.speech.SpeechSynthesisOutputFormat[response_format_parameter])
        synthesizer = azure.cognitiveservices.speech.SpeechSynthesizer(speech_config=speech_config, audio_config=None)

        default_pitch = 0
        default_rate = 1.0

        pitch = options.get('pitch', default_pitch)
        pitch_str = f'{pitch:+.0f}Hz'
        rate = options.get('rate', default_rate)
        rate_str = f'{rate:0.1f}'


        prosody_start_str = ''
        prosody_end_str = ''

        if pitch != default_pitch or rate != default_rate:
            prosody_start_str = f"""<prosody pitch="{pitch_str}" rate="{rate_str}" >"""
            prosody_end_str = """</prosody>"""

        # do some cleaning on the text
        text = text.replace(' & ', ' &amp; ')

        # the ssml str must be super optimized to have no whitespace, no extra characters
        ssml_str = f"""<speak version="1.0" xmlns="https://www.w3.org/2001/10/synthesis" xmlns:mstts="https://www.w3.org/2001/mstts" xml:lang="en-US">
<voice name="{voice_key['name']}">
{prosody_start_str}""".replace('\n', '') + text + f"""
{prosody_end_str}
</voice>
</speak>""".replace('\n', '')

        # print(f'[{ssml_str}] len: {len(ssml_str)}')

        result = synthesizer.speak_ssml(ssml_str)
        if result.reason != azure.cognitiveservices.speech.ResultReason.SynthesizingAudioCompleted:
            # special case errors:
            if 'standard voices will no longer be supported' in result.cancellation_details.error_details:
                error_message = 'Azure Standard voices are not supported anymore, please switch to Neural voices.'
            else:
                error_message = f'Could not generate audio: {result.cancellation_details.reason} {result.cancellation_details.error_details}'
            raise cloudlanguagetools.errors.RequestError(error_message)

        stream = azure.cognitiveservices.speech.AudioDataStream(result)
        stream.save_to_wav_file(output_temp_filename)

        return output_temp_file

    def get_tts_voice_list(self):
        # returns list of TtSVoice

        token = self.get_token()

        base_url = f'https://{self.region}.tts.speech.microsoft.com/'
        path = 'cognitiveservices/voices/list'
        constructed_url = base_url + path
        headers = {
            'Authorization': 'Bearer ' + token,
        }        
        response = requests.get(constructed_url, headers=headers, 
            timeout=cloudlanguagetools.constants.RequestTimeout)
        response.raise_for_status()
        voice_list = json.loads(response.content)
        result = []
        for voice_data in voice_list:
            # print(voice_data['Status'])
            try:
                result.append(AzureVoice(voice_data))
            except KeyError:
                logging.error(f'could not process voice for {voice_data}', exc_info=True)
        return result

    def get_tts_voice_list_v3(self) -> List[cloudlanguagetools.ttsvoice.TtsVoice_v3]:
        # returns list of TtsVoice_v3

        token = self.get_token()

        base_url = f'https://{self.region}.tts.speech.microsoft.com/'
        path = 'cognitiveservices/voices/list'
        constructed_url = base_url + path
        headers = {
            'Authorization': 'Bearer ' + token,
        }        
        response = requests.get(constructed_url, headers=headers, 
            timeout=cloudlanguagetools.constants.RequestTimeout)
        response.raise_for_status()
        voice_list = json.loads(response.content)
        result = []
        for voice_data in voice_list:
            # print(voice_data['Status'])
            try:
                result.append(build_tts_voice_v3(voice_data))
            except:
                logger.exception(f'could not process voice for {voice_data}')
        return result            

    def get_translation(self, text, from_language_key, to_language_key):
        base_url = f'{self.url_translator_base}/translate?api-version=3.0'
        params = f'&to={to_language_key}&from={from_language_key}'
        url = base_url + params

        # You can pass more than one object in body.
        body = [{
            'text': text
        }]
        request = requests.post(url, headers=self.get_translator_headers(), json=body, timeout=cloudlanguagetools.constants.RequestTimeout)
        response = request.json()

        if 'error' in response:
            error_message = f'Azure: could not translate text [{text}] from {from_language_key} to {to_language_key} ({response})'
            raise cloudlanguagetools.errors.RequestError(error_message)

        return response[0]['translations'][0]['text']

    def get_transliteration(self, text, transliteration_key):
        return self.transliteration(text, transliteration_key['language_id'], transliteration_key['from_script'], transliteration_key['to_script'])

    def get_translation_language_list(self):
        azure_data = self.get_supported_languages()
        result = []
        for language_id, data in azure_data['translation'].items():
            try:
                result.append(AzureTranslationLanguage(language_id))
            except KeyError:
                logging.warning(f'could not process translation language for {language_id}, {data}', exc_info=True)
        return result

    def get_transliteration_language_list(self):
        result = []
        azure_data = self.get_supported_languages()
        for language_id, data in azure_data['transliteration'].items():
            try:
                # get the first script
                for from_script_data in data['scripts']:
                    from_script =  from_script_data['code']
                    from_native_name = from_script_data['nativeName']
                    for to_script_data in from_script_data['toScripts']:
                        to_script = to_script_data['code']
                        from_script_name = from_script_data['name']
                        to_script_name = to_script_data['name']
                        to_native_name = to_script_data['nativeName']
                    result.append(AzureTransliterationLanguage(
                        language_id, 
                        from_script, 
                        to_script, 
                        from_script_name, 
                        to_script_name,
                        from_native_name,
                        to_native_name))
            except KeyError:
                logging.error(f'could not process transliteration language for {language_id}, {data}', exc_info=True)
        return result


    @cachetools.cached(cache=cachetools.TTLCache(maxsize=1024, ttl=cloudlanguagetools.constants.TTLCacheTimeout))
    def get_supported_languages(self):
        max_retries = 3
        retry_delay = 5  # seconds
        url = 'https://api.cognitive.microsofttranslator.com/languages?api-version=3.0'

        # If you encounter any issues with the base_url or path, make sure
        # that you are using the latest endpoint: https://docs.microsoft.com/azure/cognitive-services/translator/reference/v3-0-languages
        path = '/languages?api-version=3.0'

        headers = {
            'Content-type': 'application/json',
            'X-ClientTraceId': str(uuid.uuid4())
        }

        for attempt in range(max_retries):
            try:
                request = requests.get(url, headers=headers, 
                    timeout=cloudlanguagetools.constants.RequestTimeoutLong)
                request.raise_for_status()
                response = request.json()
                return response
            except requests.exceptions.RequestException as e:
                if attempt < max_retries - 1:  # If not the last attempt
                    logger.warning(f"Request failed. Retrying in {retry_delay} seconds... (Attempt {attempt + 1}/{max_retries})")
                    time.sleep(retry_delay)
                else:
                    logger.error(f"Max retries reached. Unable to get supported languages.")
                    raise  # Re-raise the last exception if all retries failed

        # print(json.dumps(response, sort_keys=True, indent=4, ensure_ascii=False, separators=(',', ': ')))        


    def detect_language(self, text_list):
        url = f'{self.url_translator_base}/detect?api-version=3.0'
        body = [{'text': text} for text in text_list]

        request = requests.post(url, headers=self.get_translator_headers(), json=body, timeout=cloudlanguagetools.constants.RequestTimeout)
        response = request.json()

        language_score = {}
        for entry in response:
            score = entry['score']
            language = entry['language']
            if language not in language_score:
                language_score[language] = 0
            language_score[language] += score

        highest_language = max(language_score.items(), key=operator.itemgetter(1))[0]
        return get_translation_language_enum(highest_language)

    def transliteration(self, text, language_key, from_script, to_script):
        url = f'{self.url_translator_base}/transliterate?api-version=3.0'
        params = f'&language={language_key}&fromScript={from_script}&toScript={to_script}'
        constructed_url = url + params

        body = [{
            'text': text
        }]
        request = requests.post(constructed_url, headers=self.get_translator_headers(), json=body, timeout=cloudlanguagetools.constants.RequestTimeout)
        response = request.json()

        assert(len(response) == 1)
        return response[0]['text']

    # supported languages: https://docs.microsoft.com/en-us/azure/cognitive-services/speech-service/language-support#speech-to-text
    def speech_to_text(self, mp3_filepath, audio_format, language=None):
        speech_config = azure.cognitiveservices.speech.SpeechConfig(subscription=self.key, region=self.region)

        if audio_format == cloudlanguagetools.options.AudioFormat.mp3:
            sound = pydub.AudioSegment.from_mp3(mp3_filepath)
        elif audio_format in [cloudlanguagetools.options.AudioFormat.ogg_opus, cloudlanguagetools.options.AudioFormat.ogg_vorbis]:
            sound = pydub.AudioSegment.from_ogg(mp3_filepath)
        elif audio_format == cloudlanguagetools.options.AudioFormat.wav:
            sound = pydub.AudioSegment.from_wav(mp3_filepath)
        wav_filepath = tempfile.NamedTemporaryFile(suffix='.wav').name
        sound.export(wav_filepath, format="wav")

        audio_input = azure.cognitiveservices.speech.audio.AudioConfig(filename=wav_filepath)

        # Creates a recognizer with the given settings
        if language != None:
            # we know which language
            logger.info(f'configuration speech recognition for language {language}')
            speech_recognizer = azure.cognitiveservices.speech.SpeechRecognizer(speech_config=speech_config, audio_config=audio_input, language=language)
        else:
            # language unknown
            logger.info(f'configuration speech recognition for any language')
            speech_recognizer = azure.cognitiveservices.speech.SpeechRecognizer(speech_config=speech_config, audio_config=audio_input)

        # Starts speech recognition, and returns after a single utterance is recognized. The end of a
        # single utterance is determined by listening for silence at the end or until a maximum of 15
        # seconds of audio is processed.  The task returns the recognition text as result. 
        # Note: Since recognize_once() returns only a single utterance, it is suitable only for single
        # shot recognition like command or query. 
        # For long-running multi-utterance recognition, use start_continuous_recognition() instead.
        result = speech_recognizer.recognize_once()

        # Checks result.
        if result.reason == azure.cognitiveservices.speech.ResultReason.RecognizedSpeech:
            return result.text
        elif result.reason == azure.cognitiveservices.speech.ResultReason.NoMatch:
            error_message = "No speech could be recognized: {}".format(result.no_match_details)
            raise Exception(error_message)
        elif result.reason == azure.cognitiveservices.speech.ResultReason.Canceled:
            cancellation_details = result.cancellation_details
            error_message = "Speech Recognition canceled: {}".format(cancellation_details)
            raise Exception(error_message)

        raise "unknown error"

    def get_dictionary_lookup_list(self):
        result = []

        azure_data = self.get_supported_languages()
        for source_language_id, data in azure_data['dictionary'].items():
            for translation in data['translations']:
                target_language_id = translation['code']
                result.append(AzureDictionaryLookup(source_language_id, 
                                target_language_id, 
                                cloudlanguagetools.constants.DictionaryLookupType.Definitions))
                result.append(AzureDictionaryLookup(source_language_id, 
                                target_language_id, 
                                cloudlanguagetools.constants.DictionaryLookupType.PartOfSpeech))
                result.append(AzureDictionaryLookup(source_language_id, 
                                target_language_id, 
                                cloudlanguagetools.constants.DictionaryLookupType.PartOfSpeechDefinitions))

        return result

    def iterate_dictionary_results(self, text, lookup_key):
        from_language_code = lookup_key['source_language_code']
        to_language_code = lookup_key['target_language_code']
        base_url = f'{self.url_translator_base}/dictionary/lookup?api-version=3.0'
        params = f'&to={to_language_code}&from={from_language_code}'
        url = base_url + params

        logger.info(f'querying url for dictionary lookup: {url}')

        # You can pass more than one object in body.
        body = [{
            'text': text
        }]
        request = requests.post(url, headers=self.get_translator_headers(), json=body, timeout=cloudlanguagetools.constants.RequestTimeout)
        response = request.json()
        if len(response) > 1:
            raise Exception(f'more than one response entries, {url}, {text}')

        translation_entries = response[0]['translations']
        for translation_entry in translation_entries:
            yield translation_entry

    def collect_definitions(self, generator):
        result = []
        for entry in generator:
            result.append(entry['displayTarget'])
        return result

    def collect_partofspeech(self, generator):
        result = []
        for entry in generator:
            result.append(entry['posTag'].lower())
        result = list(set(result))
        result.sort()
        return result

    def collect_partofspeech_definitions(self, generator):
        result = {}
        for entry in generator:
            part_of_speech = entry['posTag'].lower()
            if part_of_speech not in result:
                result[part_of_speech] = []
            result[part_of_speech].append(entry['displayTarget'])
        return result        

    def collect_result_check_empty(self, generator, collect_result_fn, text):
        result = collect_result_fn(generator)

        not_found_error_msg = f'Azure: no results found for {text}'

        if type(result) is list:
            if len(result) == 0:
                raise cloudlanguagetools.errors.NotFoundError(not_found_error_msg)
        if type(result) is dict:
            if len(result.keys()) == 0:
                raise cloudlanguagetools.errors.NotFoundError(not_found_error_msg)

        return result

    def get_dictionary_lookup(self, text, lookup_key):
        lookup_type = cloudlanguagetools.constants.DictionaryLookupType[lookup_key['lookup_type']]
        lookup_type_fn_map = {
            cloudlanguagetools.constants.DictionaryLookupType.Definitions: self.collect_definitions,
            cloudlanguagetools.constants.DictionaryLookupType.PartOfSpeech: self.collect_partofspeech,
            cloudlanguagetools.constants.DictionaryLookupType.PartOfSpeechDefinitions: self.collect_partofspeech_definitions,
        }
        lookup_fn = lookup_type_fn_map[lookup_type]
        generator = self.iterate_dictionary_results(text, lookup_key)

        return self.collect_result_check_empty(generator, lookup_fn, text)


    def custom_dictionary_lookup(self, input_text, from_language_key, to_language_key):
        base_url = f'{self.url_translator_base}/dictionary/lookup?api-version=3.0'
        params = f'&to={to_language_key}&from={from_language_key}'
        url = base_url + params

        logging.info(f'querying url for dictionary lookup: {url}')

        # You can pass more than one object in body.
        body = [{
            'text': input_text
        }]
        request = requests.post(url, headers=self.get_translator_headers(), json=body, timeout=cloudlanguagetools.constants.RequestTimeout)
        response = request.json()

        pprint.pprint(response)

        translation_entries = response[0]['translations']
        translations = [entry['displayTarget'] for entry in translation_entries]

        return translations
        
        # print(json.dumps(response, sort_keys=True, indent=4, ensure_ascii=False, separators=(',', ': ')))

        # return response[0]['translations'][0]['text']

    
    def dictionary_examples(self, input_text, translation, from_language_key, to_language_key):
        base_url = f'{self.url_translator_base}/dictionary/examples?api-version=3.0'
        params = f'&to={to_language_key}&from={from_language_key}'
        url = base_url + params

        print(url)

        # You can pass more than one object in body.
        body = [{
            'text': input_text,
            'translation': translation
        }]
        request = requests.post(url, headers=self.get_translator_headers(), json=body, timeout=cloudlanguagetools.constants.RequestTimeout)
        response = request.json()

        print(json.dumps(response, sort_keys=True, indent=4, ensure_ascii=False, separators=(',', ': ')))

        # return response[0]['translations'][0]['text']    

    



from openai import OpenAI

import logging
import tempfile
import pydub
import pprint
from typing import List

import cloudlanguagetools.service
import cloudlanguagetools.constants
import cloudlanguagetools.languages
import cloudlanguagetools.options

from cloudlanguagetools.languages import AudioLanguage
from cloudlanguagetools.options import AudioFormat

logger = logging.getLogger(__name__)

DEFAULT_TTS_SPEED = 1.0

VOICE_OPTIONS = {
            'speed' : {
                'type': cloudlanguagetools.options.ParameterType.number.name,
                'min': 0.25,
                'max': 4.0,
                'default': DEFAULT_TTS_SPEED
            },
            cloudlanguagetools.options.AUDIO_FORMAT_PARAMETER: {
                'type': cloudlanguagetools.options.ParameterType.list.name,
                'values': [
                    cloudlanguagetools.options.AudioFormat.mp3.name,
                    cloudlanguagetools.options.AudioFormat.ogg_opus.name,
                    cloudlanguagetools.options.AudioFormat.wav.name
                ],
                'default': cloudlanguagetools.options.AudioFormat.mp3.name
            }
}

TTS_SUPPORTED_LANGUAGES = [
            AudioLanguage.af_ZA,
            AudioLanguage.ar_XA,
            AudioLanguage.hy_AM,
            AudioLanguage.az_AZ,
            AudioLanguage.be_BY,
            AudioLanguage.bs_BA,
            AudioLanguage.bg_BG,
            AudioLanguage.ca_ES,
            AudioLanguage.zh_CN,
            AudioLanguage.hr_HR,
            AudioLanguage.cs_CZ,
            AudioLanguage.da_DK,
            AudioLanguage.nl_NL,
            AudioLanguage.en_US,
            AudioLanguage.et_EE,
            AudioLanguage.fi_FI, 
            AudioLanguage.fr_FR, 
            AudioLanguage.gl_ES, 
            AudioLanguage.de_DE, 
            AudioLanguage.el_GR, 
            AudioLanguage.he_IL, 
            AudioLanguage.hi_IN, 
            AudioLanguage.hu_HU, 
            AudioLanguage.is_IS, 
            AudioLanguage.id_ID, 
            AudioLanguage.it_IT, 
            AudioLanguage.ja_JP, 
            AudioLanguage.kn_IN, 
            AudioLanguage.kk_KZ, 
            AudioLanguage.ko_KR, 
            AudioLanguage.lv_LV, 
            AudioLanguage.lt_LT,
            AudioLanguage.mk_MK,
            AudioLanguage.ms_MY,
            AudioLanguage.mr_IN,
            AudioLanguage.ne_NP,
            AudioLanguage.nb_NO,
            AudioLanguage.fa_IR,
            AudioLanguage.pl_PL,
            AudioLanguage.pt_PT,
            AudioLanguage.pt_BR,
            AudioLanguage.ro_RO,
            AudioLanguage.ru_RU,
            AudioLanguage.sr_RS,
            AudioLanguage.sk_SK,
            AudioLanguage.sl_SI,
            AudioLanguage.es_ES,
            AudioLanguage.es_MX,
            AudioLanguage.sw_KE,
            AudioLanguage.sv_SE,
            # AudioLanguage.tl_PH, # need to add tagalog language in AudioLanguages
            AudioLanguage.ta_IN,
            AudioLanguage.ta_LK,
            AudioLanguage.th_TH,
            AudioLanguage.tr_TR,
            AudioLanguage.uk_UA,
            AudioLanguage.ur_PK,
            AudioLanguage.ur_IN,
            AudioLanguage.vi_VN,
            AudioLanguage.cy_GB            
]

class OpenAIVoice(cloudlanguagetools.ttsvoice.TtsVoice):
    def __init__(self, name: str, 
                audio_language: cloudlanguagetools.languages.AudioLanguage,
                gender: cloudlanguagetools.constants.Gender):
        self.name = name
        self.gender = gender
        self.audio_language = audio_language
        self.service = cloudlanguagetools.constants.Service.OpenAI
        self.service_fee = cloudlanguagetools.constants.ServiceFee.paid

    def get_voice_key(self):
        return {
            'name': self.name,
            'language': self.audio_language.name            
        }

    def get_voice_shortname(self):
        return self.name

    def get_options(self):
        return VOICE_OPTIONS

class OpenAIService(cloudlanguagetools.service.Service):
    def __init__(self):
        self.chatbot_model = "gpt-3.5-turbo"

    def configure(self, config):
        self.api_key = config['api_key']
        self.client = OpenAI(api_key=self.api_key)

    def single_prompt(self, prompt, max_tokens):
        messages = [
            {'role': 'user', 'content': prompt}
        ]        

        if max_tokens != None:
            response = self.client.chat.completions.create(model=self.chatbot_model,
            messages=messages,
            max_tokens=max_tokens)    
        else:
            response = self.client.chat.completions.create(model=self.chatbot_model,
            messages=messages)    
        logger.debug(pprint.pformat(response))
        tokens_used = response.usage.total_tokens
        response_text = response.choices[0].message.content
        return response_text, tokens_used

    def full_query(self, messages, max_tokens):
        if max_tokens != None:
            response = self.client.chat.completions.create(model=self.chatbot_model,
            messages=messages,
            max_tokens=max_tokens)    
        else:
            response = self.client.chat.completions.create(model=self.chatbot_model,
            messages=messages)    
        logger.debug(pprint.pformat(response))
        return response

    def speech_to_text(self, filepath, audio_format: cloudlanguagetools.options.AudioFormat):

        if audio_format in [cloudlanguagetools.options.AudioFormat.ogg_opus, cloudlanguagetools.options.AudioFormat.ogg_vorbis]:
            # need to convert to wav first
            sound = pydub.AudioSegment.from_ogg(filepath)
            wav_tempfile = tempfile.NamedTemporaryFile(prefix='cloudlanguagetools_OpenAI_speech_to_text', suffix='.wav')
            sound.export(wav_tempfile.name, format="wav")
            filepath = wav_tempfile.name

        logger.debug(f'opening file {filepath}')
        audio_file= open(filepath, "rb")
        transcript = self.client.audio.transcriptions.create(model="whisper-1", file=audio_file)
        return transcript.text
    
    
    def get_tts_voice_list(self) -> List[OpenAIVoice]:
        result = []

        for audio_language in TTS_SUPPORTED_LANGUAGES:
            result.extend([
                OpenAIVoice('alloy', audio_language, cloudlanguagetools.constants.Gender.Female),
                OpenAIVoice('echo', audio_language, cloudlanguagetools.constants.Gender.Male),
                OpenAIVoice('fable', audio_language, cloudlanguagetools.constants.Gender.Female),
                OpenAIVoice('onyx', audio_language, cloudlanguagetools.constants.Gender.Male),
                OpenAIVoice('nova', audio_language, cloudlanguagetools.constants.Gender.Female),
                OpenAIVoice('shimmer', audio_language, cloudlanguagetools.constants.Gender.Female),
            ])
        return result

    def build_tts_voice_v3(self, voice_name, gender):
        return cloudlanguagetools.ttsvoice.TtsVoice_v3(
            name=voice_name,
            voice_key={
                'name': voice_name
            },
            options=VOICE_OPTIONS,
            service=cloudlanguagetools.constants.Service.OpenAI,
            gender=gender,
            audio_languages=TTS_SUPPORTED_LANGUAGES,
            service_fee=cloudlanguagetools.constants.ServiceFee.paid
        )

    def get_tts_voice_list_v3(self) -> List[cloudlanguagetools.ttsvoice.TtsVoice_v3]:
        # returns list of TtsVoice_v3

        result = [
            self.build_tts_voice_v3('alloy', cloudlanguagetools.constants.Gender.Female),
            self.build_tts_voice_v3('echo', cloudlanguagetools.constants.Gender.Male),
            self.build_tts_voice_v3('fable', cloudlanguagetools.constants.Gender.Female),
            self.build_tts_voice_v3('onyx', cloudlanguagetools.constants.Gender.Male),
            self.build_tts_voice_v3('nova', cloudlanguagetools.constants.Gender.Female),
            self.build_tts_voice_v3('shimmer', cloudlanguagetools.constants.Gender.Female)
        ]
        return result

    def get_tts_audio(self, text, voice_key, options):
        # https://platform.openai.com/docs/guides/text-to-speech
        # https://platform.openai.com/docs/api-reference/audio/createSpeech?lang=python
        
        output_temp_file = tempfile.NamedTemporaryFile()

        speed = options.get('speed', DEFAULT_TTS_SPEED)
        response_format_parameter, audio_format = self.get_request_audio_format({
            AudioFormat.mp3: 'mp3',
            AudioFormat.ogg_opus: 'opus',
            AudioFormat.wav: 'wav'
        }, options, AudioFormat.mp3)

        response = self.client.audio.speech.create(
            model='tts-1-hd',
            voice=voice_key['name'],
            input=text,
            response_format=response_format_parameter,
            speed=speed
        )
        response.stream_to_file(output_temp_file.name)

        return output_temp_file



import os
import sys
import logging
import unittest
import pytest
import json
import pprint

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import cloudlanguagetools
import cloudlanguagetools.servicemanager
import cloudlanguagetools.constants
from cloudlanguagetools.languages import Language
from cloudlanguagetools.constants import Service
import cloudlanguagetools.errors

def get_manager():
    manager = cloudlanguagetools.servicemanager.ServiceManager()
    manager.configure_default()
    return manager

class TestTranslation(unittest.TestCase):
    def setUp(self):
        self.manager = get_manager()

    def test_service_cost(self):
        self.assertEqual(self.manager.service_cost('abcd', 'Azure', cloudlanguagetools.constants.RequestType.translation), 4)
        self.assertEqual(self.manager.service_cost('abcd', 'Epitran', cloudlanguagetools.constants.RequestType.transliteration), 0)
        self.assertEqual(self.manager.service_cost('abcd', 'Naver', cloudlanguagetools.constants.RequestType.translation), 4)
        self.assertEqual(self.manager.service_cost('abcd', 'Naver', cloudlanguagetools.constants.RequestType.audio), 24)
        self.assertEqual(self.manager.service_cost('abcd', 'MandarinCantonese', cloudlanguagetools.constants.RequestType.transliteration), 0)


    def test_language_data_json_v2(self):
        language_data = self.manager.get_language_data_json_v2()
        
        # check free translation services
        free_translation_options = language_data['free']['translation_options']
        # check that libretranslate is there
        free_translation_services = list(set([option['service'] for option in free_translation_options]))
        self.assertEqual(free_translation_services, []) # libretranslate removed

        # check free transliteration services
        free_transliteration_options = language_data['free']['transliteration_options']
        free_transliteration_services = list(set([option['service'] for option in free_transliteration_options]))
        free_transliteration_services.sort()
        self.assertEqual(free_transliteration_services, ['Epitran', 'MandarinCantonese', 'PyThaiNLP'])